/*****************************************************************************
 *
 *     Author:	Frederick Vong
 *
 * Modification Log:
 * -----------------
 * .01	06-01-94	vong	initial version
 *
 ***************************************************************************/
/*
*****************************************************************
                          COPYRIGHT NOTIFICATION
*****************************************************************

THE FOLLOWING IS A NOTICE OF COPYRIGHT, AVAILABILITY OF THE CODE,
AND DISCLAIMER WHICH MUST BE INCLUDED IN THE PROLOGUE OF THE CODE
AND IN ALL SOURCE LISTINGS OF THE CODE.
 
(C)  COPYRIGHT 1993 UNIVERSITY OF CHICAGO
 
Argonne National Laboratory (ANL), with facilities in the States of 
Illinois and Idaho, is owned by the United States Government, and
operated by the University of Chicago under provision of a contract
with the Department of Energy.

Portions of this material resulted from work developed under a U.S.
Government contract and are subject to the following license:  For
a period of five years from March 30, 1993, the Government is
granted for itself and others acting on its behalf a paid-up,
nonexclusive, irrevocable worldwide license in this computer
software to reproduce, prepare derivative works, and perform
publicly and display publicly.  With the approval of DOE, this
period may be renewed for two additional five year periods. 
Following the expiration of this period or periods, the Government
is granted for itself and others acting on its behalf, a paid-up,
nonexclusive, irrevocable worldwide license in this computer
software to reproduce, prepare derivative works, distribute copies
to the public, perform publicly and display publicly, and to permit
others to do so.

*****************************************************************
                                DISCLAIMER
*****************************************************************

NEITHER THE UNITED STATES GOVERNMENT NOR ANY AGENCY THEREOF, NOR
THE UNIVERSITY OF CHICAGO, NOR ANY OF THEIR EMPLOYEES OR OFFICERS,
MAKES ANY WARRANTY, EXPRESS OR IMPLIED, OR ASSUMES ANY LEGAL
LIABILITY OR RESPONSIBILITY FOR THE ACCURACY, COMPLETENESS, OR
USEFULNESS OF ANY INFORMATION, APPARATUS, PRODUCT, OR PROCESS
DISCLOSED, OR REPRESENTS THAT ITS USE WOULD NOT INFRINGE PRIVATELY
OWNED RIGHTS.  

*****************************************************************
LICENSING INQUIRIES MAY BE DIRECTED TO THE INDUSTRIAL TECHNOLOGY
DEVELOPMENT CENTER AT ARGONNE NATIONAL LABORATORY (708-252-2000).
*/

#include "probe.h"
#include <iomanip.h>

// this must be defined as same
// as in the "alarm.h"

char *severityStrs[] = {
      "No Alarm",
      "Minor   ",
      "Major   ",
      "Invalid ",
      "Unknown ",
};

char *statusStrs[] = {
      "No Alarm ",
      "Read     ",
      "Write    ",
      "HiHi     ",
      "High     ",
      "LoLo     ",
      "Low      ",
      "State    ",
      "COS      ",
      "Comm     ",
      "Timeout  ",
      "HW limit ",
      "Calc     ",
      "Scan     ",
      "Link     ",
      "Soft     ",
      "Bad Sub  ",
      "UDF      ",
      "Disable  ",
      "Simm     ",
      "RD Access",
      "WR Access",
      "Unknown  ",
      "Unknown  ",
      "Unknown  ",
      "Unknown  ",
      "Unknown  ",
      "Unknown  ",
      "Unknown  ",
      "Unknown  ",
      "Unknown  ",
      "Unknown  ",
};

static char *buttonLabels[16] = {
	   "0",
	   "1",
	   "2",
	   "3",
	   "4",
	   "5",
	   "6",
	   "7",
	   "8",
	   "9",
	   "10",
	   "11",
	   "12",
	   "13",
	   "14",
	   "15",};

inline void writeLabel(Widget w, char * str) {
  XmString xmstr = XmStringLtoRCreate(str, XmSTRING_DEFAULT_CHARSET);
  XtVaSetValues(w, XmNlabelString, xmstr, NULL);
  XmStringFree(xmstr);
}

/* returns a copy of the name with .RTYP appended to it */
/* name may not have two '.' as .RTYP is appended after the first '.' */
char *setName(const char *name) {
  int dot = 0;
  size_t len;
  char *type = ".RTYP\0";
  char *newName;
  /* remove any leading white space */
  while (*name == ' ') {
    name += 1;
  }
  if (name != '\0') {
    len = strlen(name);
  }
  else { len = 0; }

  for (; dot < len; dot += 1) {
    if ('.' == name[dot]) {
      break;
    }
  }
  if (dot == len) {
    newName = new char[len+6];
    memcpy(newName, name, len);
    memcpy(newName + len, type, 6);
  }
  else {
    newName = new char[dot+6];
    memcpy(newName, name, dot);
    memcpy(newName + dot, type, 6);
  }
  return newName;
}


myCh::myCh(Widget parent, const char *name, vCh *recCh) : vCh(name) {


  // load fonts
  _font1 = XLoadQueryFont(XtDisplay(parent),
     "-adobe-times-bold-i-normal--18-180-75-75-p-98-iso8859-1");
  if (_font1) {
    _fontList1 = XmFontListCreate(_font1, XmSTRING_DEFAULT_CHARSET);
  } else {
    _fontList1 = NULL;
  }
    
  _font2 = XLoadQueryFont(XtDisplay(parent),
     "-adobe-helvetica-bold-o-normal--24-240-75-75-p-138-iso8859-1");
  if (_font2) {
    _fontList2 = XmFontListCreate(_font2, XmSTRING_DEFAULT_CHARSET);
  } else {
    _fontList2 = NULL;
  }

  _font3 = XLoadQueryFont(XtDisplay(parent), "fixed");
  if (_font3) {
    _fontList3 = XmFontListCreate(_font3, XmSTRING_DEFAULT_CHARSET);
  } else {
    _fontList3 = NULL;
  }

  _panel = XtVaCreateWidget("panel",
    xmPanedWindowWidgetClass,parent,
    XmNsashHeight, 1,
    XmNsashWidth, 1,
    NULL);
  _rowCol = XtVaCreateWidget("rowCol",
    xmRowColumnWidgetClass,_panel,NULL);
 
  // create a label widget for channel name
  if (_font1) {
    _nameLabel = XtVaCreateManagedWidget("Channel name",
              xmLabelWidgetClass,_rowCol,
	      XmNfontList, _fontList1,
	      NULL);
  } else {
    _nameLabel = XtVaCreateManagedWidget("Channel name",
              xmLabelWidgetClass,_rowCol,NULL);
  }

  // create a label widget for channel value
  if (_font2) {
    _value = XtVaCreateManagedWidget("Channel value",
              xmLabelWidgetClass,_rowCol,
	      XmNfontList, _fontList2,
	      NULL);
  } else {
    _value = XtVaCreateManagedWidget("Channel value",
              xmLabelWidgetClass,_rowCol,
              NULL);
  }

  // create a label for the status field
  _status = XtVaCreateManagedWidget("status",
    xmLabelWidgetClass,_rowCol,NULL);

  // create a text entry for the channel name input
  if (_font1) {
    _nameEntry = XtVaCreateManagedWidget("Channel value",
              xmTextFieldWidgetClass,_panel,
	      XmNfontList, _fontList1,
	      NULL);
  } else {
    _nameEntry = XtVaCreateManagedWidget("Channel value",
              xmTextFieldWidgetClass,_panel,NULL);
  }
  XtAddCallback(_nameEntry,XmNactivateCallback,myCh::nameEntryCb,this);
  XtAddCallback(_nameEntry,XmNvalueChangedCallback,myCh::nameEntryValueChangedCb,this);

  XtManageChild(_rowCol);

  // create a rowColumn widget to hold the buttons
  _buttonsHolder = XtVaCreateWidget("buttonsHolder",
	      xmRowColumnWidgetClass,_panel,
	      XmNorientation, XmHORIZONTAL,
	      XmNnumColumns, 2,
              XmNpacking, XmPACK_COLUMN,
	      NULL);

  // create a pushbutton widget for invoke the connect function
  if (_font1) {
    _connect = XtVaCreateManagedWidget("Connect",
              xmPushButtonWidgetClass,_buttonsHolder,
	      XmNfontList, _fontList1,
	      NULL);
  } else {
    _connect = XtVaCreateManagedWidget("Connect",
              xmPushButtonWidgetClass,_buttonsHolder,NULL);
  }

  XtAddCallback(_connect,XmNactivateCallback,myCh::nameEntryCb,this);

  // create a pushbutton widget for invoke the adjust dialog
  if (_font1) {
    _adjust = XtVaCreateManagedWidget("Adjust",
              xmPushButtonWidgetClass,_buttonsHolder,
	      XmNfontList, _fontList1,
	      NULL);
  } else {
    _adjust = XtVaCreateManagedWidget("Adjust",
              xmPushButtonWidgetClass,_buttonsHolder);
  }
  XtAddCallback(_adjust,XmNactivateCallback,myCh::adjustCb,this);

  // create a pushbutton widget for invoke the info dialog
  if (_font1) {
    _info = XtVaCreateManagedWidget("Info",
              xmPushButtonWidgetClass,_buttonsHolder,
	      XmNfontList, _fontList1,
	      NULL);
  } else {
    _info = XtVaCreateManagedWidget("Info",
              xmPushButtonWidgetClass,_buttonsHolder,NULL);
  }
  XtAddCallback(_info,XmNactivateCallback,myCh::infoCb,this);

  // create a pushbutton widget for invoke the format dialog
  if (_font1) {
    _format = XtVaCreateManagedWidget("Format",
              xmPushButtonWidgetClass,_buttonsHolder,
	      XmNfontList, _fontList1,
	      NULL);
  } else {
    _format = XtVaCreateManagedWidget("Format",
              xmPushButtonWidgetClass,_buttonsHolder,NULL);
  }
  XtAddCallback(_format,XmNactivateCallback,myCh::formatCb,this);

  // create a pushbutton widget for invoke the version function
  if (_font1) {
    _help = XtVaCreateManagedWidget("Version",
              xmPushButtonWidgetClass,_buttonsHolder,
	      XmNfontList, _fontList1,
	      NULL);
  } else {
    _help = XtVaCreateManagedWidget("Version",
              xmPushButtonWidgetClass,_buttonsHolder,NULL);
  }
  XtAddCallback(_help,XmNactivateCallback,myCh::helpCb,this);

  // create a pushbutton widget for invoke the quit function
  if (_font1) {
    _quit = XtVaCreateManagedWidget("Quit",
              xmPushButtonWidgetClass,_buttonsHolder,
	      XmNfontList, _fontList1,
	      NULL);
  } else {
    _quit = XtVaCreateManagedWidget("Quit",
              xmPushButtonWidgetClass,_buttonsHolder,NULL);
  }
  XtAddCallback(_quit,XmNactivateCallback,myCh::quitCb,this);

  XtManageChild(_buttonsHolder);

  XtManageChild(_panel);

  XtSetSensitive(_connect, False);
  XtSetSensitive(_adjust, False);
  XtSetSensitive(_info, False);
  XtSetSensitive(_format, False);
  // fixed the size of the display
/*
  Dimension h, w;
  XtVaGetValues(_rowCol,
	      XmNheight, &h,
	      NULL);

  XtVaSetValues(_rowCol,
	      XmNpaneMaximum, h,
	      XmNpaneMinimum, h,
	      NULL);

  XtVaGetValues(_buttonsHolder,
	      XmNheight, &h,
	      NULL);

  XtVaSetValues(_buttonsHolder,
	      XmNpaneMaximum, h,
	      XmNpaneMinimum, h,
	      NULL);
*/
  // initialize the following widgets
  _adjustDlg = NULL;
  _infoDlg = NULL;
  _formatDlg = NULL;
  _userDlg = NULL;
  _helpDlg = NULL;

  // initialize the following strings
  _userName = NULL;
  _hostName = NULL;

  /* create a vCh used to access the record type */
  RvCh = recCh;
  return;
}

myCh::~myCh() {
  if (_userName) {
    delete _userName;
  }
  if (_hostName) {
    delete _hostName;
  }
  if (_panel) {
    XtDestroyWidget(_panel);
  }
}

char *myCh::name(char *newName) {
  writeLabel(_nameLabel,newName);
  writeLabel(_value,"Connecting ...");

  char * name; 
  name = setName(newName);
  RvCh->name(name);

  return vCh::name(newName);
}

void myCh::valueChangedCb() {
  char str[80];
  *this >> str;

  writeLabel(_value,str);
  if (_startTime == 0) { 
    // first time monitoring
    _startTime = pchData->_time;
    _lastMaxTime = pchData->_time;
    _lastMinTime = pchData->_time;
    _maxValue = *(pchData->_pdata);
    _minValue = *(pchData->_pdata);
  } else {
    switch (ca_field_type(pchData->_chID)) {
    case DBF_STRING :
    case DBF_ENUM :
      _lastMinTime = _lastMaxTime;
      _minValue = _maxValue;
      _lastMaxTime = pchData->_time;
      _maxValue = *(pchData->_pdata);
      if (_infoDlg)
        updateInfoHistoryPanel();
      if (_adjustDlg) {
        XmTextFieldSetString(_adjustTextEntry,stringValue());
      }
      break;
    case DBF_CHAR :
      if (_minValue.c.value > pchData->_pdata->c.value) {
        _minValue = *(pchData->_pdata);
        _lastMinTime = pchData->_time;
        if (_infoDlg)
          updateInfoHistoryPanel();
      } else 
      if (_maxValue.c.value < pchData->_pdata->c.value) {
        _maxValue = *(pchData->_pdata);
        _lastMaxTime = pchData->_time;
        if (_infoDlg)
          updateInfoHistoryPanel();
      } 
      if (_adjustDlg) {
        XmTextFieldSetString(_adjustTextEntry,stringValue());
        updateScrollBar();
      }
      break;
    case DBF_INT :
      if (_minValue.i.value > pchData->_pdata->i.value) {
        _minValue = *(pchData->_pdata);
        _lastMinTime = pchData->_time;
        if (_infoDlg)
          updateInfoHistoryPanel();
      } else 
      if (_maxValue.i.value < pchData->_pdata->i.value) {
        _maxValue = *(pchData->_pdata);
        _lastMaxTime = pchData->_time;
        if (_infoDlg)
          updateInfoHistoryPanel();
      } 
      if (_adjustDlg) {
        XmTextFieldSetString(_adjustTextEntry,stringValue());
        updateScrollBar();
      }
      break;
    case DBF_LONG :
      if (_minValue.l.value > pchData->_pdata->l.value) {
        _minValue = *(pchData->_pdata);
        _lastMinTime = pchData->_time;
        if (_infoDlg)
          updateInfoHistoryPanel();
      } else 
      if (_maxValue.l.value < pchData->_pdata->l.value) {
        _maxValue = *(pchData->_pdata);
        _lastMaxTime = pchData->_time;
        if (_infoDlg)
          updateInfoHistoryPanel();
      } 
      if (_adjustDlg) {
        XmTextFieldSetString(_adjustTextEntry,stringValue());
        updateScrollBar();
      }
      break;
    case DBF_FLOAT :
      if (_minValue.f.value > pchData->_pdata->f.value) {
        _minValue = *(pchData->_pdata);
        _lastMinTime = pchData->_time;
        if (_infoDlg)
          updateInfoHistoryPanel();
      } else 
      if (_maxValue.f.value < pchData->_pdata->f.value) {
        _maxValue = *(pchData->_pdata);
        _lastMaxTime = pchData->_time;
        if (_infoDlg)
        updateInfoHistoryPanel();
      } 
      if (_adjustDlg) {
        XmTextFieldSetString(_adjustTextEntry,stringValue());
        updateScrollBar();
      }
      break;
    case DBF_DOUBLE :
      if (_minValue.d.value > pchData->_pdata->d.value) {
        _minValue = *(pchData->_pdata);
        _lastMinTime = pchData->_time;
        if (_infoDlg)
          updateInfoHistoryPanel();
      } else 
      if (_maxValue.d.value < pchData->_pdata->d.value) {
        _maxValue = *(pchData->_pdata);
        _lastMaxTime = pchData->_time;
        if (_infoDlg)
        updateInfoHistoryPanel();
      } 
      if (_adjustDlg) {
        XmTextFieldSetString(_adjustTextEntry,stringValue());
        updateScrollBar();
      }
      break;
    default :
      break;
    }
  }
  return;
}

void myCh::alarmChangedCb() {
  char str[80];
  ostrstream ost(str,80);

  // update the status

  if (ca_read_access(pchData->_chID)) {
    int status;
    int severity;
    if (pchData->_pdata == NULL) return;
    status = pchData->_pdata->d.status;
    severity = pchData->_pdata->d.severity;
    if ((status > ALARM_NSTATUS) || (status < NO_ALARM)) {
      ost << "unknown   unknown" << ends;
    } else 
    if ((severity > ALARM_NSEV) || (severity < NO_ALARM)) {
      ost << "unknown   unknown" << ends;
    } else {
      ost.width(8);
      if (ca_write_access(pchData->_chID)) {
        ost << severityStrs[severity] << "  " << statusStrs[status] << ends;
      } else {
        ost << severityStrs[severity] << "  " << statusStrs[status] << "(Read Only)" << ends;
      }
    }
  } else {
    ost << "Unknown   Unknown   (No Access)" << ends;
  }
  writeLabel(_status,str);
}

void myCh::connectCb() {
//  cout << "connectCb() : " << this->name() << " is connected!" << endl;
  writeLabel(_nameLabel,_name);
  // fill in the name in the _nameEntry
  XmTextFieldSetString(_nameEntry,_name);
  if (_formatDlg) {
    updateFormatPanel();
  }
  if (_infoDlg) {
    updateInfoLimitsPanel();
    updateInfoHistoryPanel();
  }
  // signal valueChangedCb() a new channel is connected
  _startTime = 0;
  // update the scroll bar info if necessary
  updateAdjustParameters();
  XtSetSensitive(_info,True);
  XtSetSensitive(_format,True);
  if (_adjustDlg) {
    XtSetSensitive(_adjustTextEntry,True);
  }
  return;
}

void myCh::disconnectCb() {
  writeLabel(_value,"Lost Connection ...");
  if (_adjustDlg) {
    XtSetSensitive(_adjustControlPanel,False);
  }
  return;
}

void myCh::accessRightsChangedCb() {
  _startTime = 0;
  if (_infoDlg) {
    updateInfoLimitsPanel();
  }
  if (ca_write_access(pchData->_chID)) {
    XtSetSensitive(_adjust,True);
    if (_adjustDlg) {
      XtSetSensitive(_adjustControlPanel,True);
    }
  } else {
    XtSetSensitive(_adjust,False);
    if (_adjustDlg) {
      XtSetSensitive(_adjustControlPanel,False);
    }
  }
  alarmChangedCb();
  return;
}

void myCh::nameEntryCb(Widget, XtPointer data, XtPointer) {
  myCh *pch = (myCh *) data;
  char *text = XmTextFieldGetString(pch->_nameEntry);
  char *name = text;
  char *tmp;

  /* skip the leading space */
  while (*name == ' ') {
    name++;
  }
  if (tmp = strstr(name,"\n")) {
    *tmp = '\0';
  }
 
  XmTextFieldSetString(pch->_nameEntry,name);

  // close the adjust dialog
  if (pch->_adjustDlg) {
    pch->adjustCloseButtonCb((Widget) NULL, (XtPointer) pch, (XtPointer) NULL);
  }
  // close the info dialog
  if (pch->_infoDlg) {
    pch->infoCloseButtonCb((Widget) NULL, (XtPointer) pch, (XtPointer) NULL);
  }
  if (strcmp(pch->_name,name) != 0) {
    pch->name(name);
    XtSetSensitive(pch->_info,False);
    XtSetSensitive(pch->_format,False);
    XtSetSensitive(pch->_adjust,False);
  }
  XtFree(text);
  return;
}

void myCh::nameEntryValueChangedCb(Widget w, XtPointer data, XtPointer) {
  myCh *pch = (myCh *) data;
  char *tmp = XmTextFieldGetString(w);
  if (strlen(tmp) > 0) {
    XtSetSensitive(pch->_connect,True);
  } else {
    XtSetSensitive(pch->_connect,False);
  }
  XtFree(tmp);
}

void myCh::updateFormatPanel() {
  int i;
  for (i=0; i < 16; i++) { 
    if (i == _precision) {
      XtVaSetValues(_formatPrecisionButton[i],
	  XmNset,TRUE,
	  NULL);
    } else {
      XtVaSetValues(_formatPrecisionButton[i],
	  XmNset,FALSE,
	  NULL);
    }
  }

  // set the default notation format
  switch (_notation) {
  case 'f' :
    XtVaSetValues(_formatFixedButton,
	  XmNset,TRUE,
	  NULL);
    XtVaSetValues(_formatScientificButton,
	  XmNset,FALSE,
	  NULL);
    XtVaSetValues(_formatHexadecimalButton,
	  XmNset,FALSE,
	  NULL);
    break; 
  case 'e' :
    XtVaSetValues(_formatFixedButton,
	  XmNset,FALSE,
	  NULL);
    XtVaSetValues(_formatScientificButton,
	  XmNset,TRUE,
	  NULL);
    XtVaSetValues(_formatHexadecimalButton,
	  XmNset,FALSE,
	  NULL);
    break; 
  case 'x' :
    XtVaSetValues(_formatFixedButton,
	  XmNset,FALSE,
	  NULL);
    XtVaSetValues(_formatScientificButton,
	  XmNset,FALSE,
	  NULL);
    XtVaSetValues(_formatHexadecimalButton,
  	  XmNset,TRUE,
          NULL);
    break; 
  default  :
    XtVaSetValues(_formatFixedButton,
          XmNset,TRUE,
          NULL);
    break;
  }

}

void myCh::formatCb(Widget widget, XtPointer data, XtPointer) {
  myCh *pch = (myCh *) data;
  int i;
  Position x, y;
  Dimension w, h;
  
  // see whether the dialog is created before
  if (pch->_formatDlg == NULL) {

    // get the dimension and position of the widget
    XtVaGetValues(topShell(widget),
            XmNwidth, &w,
            XmNheight, &h,
	    XmNx, &x,
	    XmNy, &y,
            NULL);
   
    // create the dialog shell
    pch->_formatDlg = new vDialog(widget,"format");
    pch->_formatDlg->moveTo(x + w + 15, y - 25);

    // create a rowColumn widget to hold both precision & notation panels
    pch->_formatFormatPanel = 
	      XtVaCreateManagedWidget("precisionPanel",
		  xmRowColumnWidgetClass,pch->_formatDlg->panel(),
		  NULL);

    // create a vFrame to hold the precision panel
    pch->_formatPrecisionPanel = new vFrame(pch->_formatFormatPanel,
				 "Precision");

    // create a radio box to hold 16 buttons
    pch->_formatPrecisionButtonPanel =
	      XmCreateRadioBox(pch->_formatPrecisionPanel->frame(),
		  "precisionButtonPanel",NULL, 0);
    XtVaSetValues(pch->_formatPrecisionButtonPanel,
		  XmNentryClass, xmToggleButtonWidgetClass,
		  XmNradioAlwaysOne, TRUE,
		  XmNnumColumns, 4,
		  XmNchildType, XmFRAME_WORKAREA_CHILD,
		  NULL);

    // create 16 buttons for the above radio box
    for (i=0;i<16;i++) {
      pch->_formatPrecisionButton[i] = 
           XmCreateToggleButton(pch->_formatPrecisionButtonPanel,
	   buttonLabels[i],NULL,0);
      XtAddCallback(pch->_formatPrecisionButton[i], XmNvalueChangedCallback,
		  myCh::formatPrecisionButtonCb,pch);
    }

    // create a frame widget to hold the notation panel
    pch->_formatNotationPanel = new vFrame(pch->_formatFormatPanel,
				  "Notation");

    // create a radio box to hold the data format
    pch->_formatNotationButtonPanel =
	      XmCreateRadioBox(pch->_formatNotationPanel->frame(),
		  "precisionButtonPanel",NULL, 0);
    XtVaSetValues(pch->_formatNotationButtonPanel,
		  XmNentryClass, xmToggleButtonWidgetClass,
		  XmNradioAlwaysOne, TRUE,
		  XmNchildType, XmFRAME_WORKAREA_CHILD,
		  NULL);

    // create 3 buttons for the above radio box
    pch->_formatFixedButton =
           XmCreateToggleButton(pch->_formatNotationButtonPanel,
	   "Fixed",NULL,0);
    XtAddCallback(pch->_formatFixedButton, XmNvalueChangedCallback,
		  myCh::formatNotationButtonCb,pch);

    pch->_formatScientificButton =
           XmCreateToggleButton(pch->_formatNotationButtonPanel,
	   "Scientific",NULL,0);
    XtAddCallback(pch->_formatScientificButton, XmNvalueChangedCallback,
		  myCh::formatNotationButtonCb,pch);

    pch->_formatHexadecimalButton =
           XmCreateToggleButton(pch->_formatNotationButtonPanel,
	   "Hexadecimal",NULL,0);
    XtAddCallback(pch->_formatHexadecimalButton, XmNvalueChangedCallback,
		  myCh::formatNotationButtonCb,pch);

    XtManageChildren(pch->_formatPrecisionButton,16);
    XtManageChild(pch->_formatFixedButton);
    XtManageChild(pch->_formatHexadecimalButton);
    XtManageChild(pch->_formatScientificButton);
    XtManageChild(pch->_formatPrecisionButtonPanel);
    XtManageChild(pch->_formatNotationButtonPanel);
    pch->_formatPrecisionPanel->manage();
    pch->_formatNotationPanel->manage();

    // create the close button and its panel
    pch->_formatActionPanel = XtVaCreateWidget("CloseButtonPanel",
	    xmFormWidgetClass, pch->_formatDlg->panel(),
	    XmNfractionBase, 3,
	    NULL);

    pch->_formatCloseButton = XtVaCreateManagedWidget("Close",
	    xmPushButtonWidgetClass, pch->_formatActionPanel,
	    XmNtopAttachment,    XmATTACH_FORM,
	    XmNbottomAttachment, XmATTACH_FORM,
	    XmNleftAttachment,   XmATTACH_POSITION,
	    XmNleftPosition,     1,
	    XmNrightAttachment,  XmATTACH_POSITION,
	    XmNrightPosition,    2,
	    NULL);

    XtAddCallback(pch->_formatCloseButton, XmNactivateCallback, 
		  myCh::formatCloseCb, pch);

    XtManageChild(pch->_formatActionPanel);
    // fix the size of the panel and the shell
    XtVaGetValues(pch->_formatCloseButton, XmNheight, &h, NULL);
    XtVaSetValues(pch->_formatActionPanel,
	    XmNpaneMaximum, h,
	    XmNpaneMinimum, h,
	    NULL);
    pch->_formatDlg->manage();

    // set the default precision and notation
    pch->updateFormatPanel();
  }
  
  pch->_formatDlg->manage();
  return;
}

void myCh::quitCb(Widget w, XtPointer data, XtPointer) {
  myCh *pch = (myCh *) data;
  if (pch->pchData != NULL) {
    pch->pchData->self->remove(*pch);
  }
  XtCloseDisplay(XtDisplay(w));
  pch->_panel = NULL;
  exit(0);
}

void myCh::formatCloseCb(Widget , XtPointer data, XtPointer) {
  myCh *pch = (myCh *) data;
  if (pch->_formatDlg)
    pch->_formatDlg->unmanage();
}

void myCh::formatPrecisionButtonCb(Widget w, XtPointer data, XtPointer cbs ) {
  XmToggleButtonCallbackStruct *pcbs = (XmToggleButtonCallbackStruct *) cbs;
  if (!pcbs->set) return;
  myCh *pch = (myCh *) data;
  int i;
  for (i=0;i<16;i++) {
    if (pch->_formatPrecisionButton[i] == w) {
      pch->precision(i);
      pch->valueChangedCb();
    }
  }
  if (pch->_infoDlg) {
    pch->updateInfoLimitsPanel();
    pch->updateInfoHistoryPanel();
  }
  return;
}

void myCh::formatNotationButtonCb(Widget w, XtPointer data, XtPointer cbs) {
  XmToggleButtonCallbackStruct *pcbs = (XmToggleButtonCallbackStruct *) cbs;
  if (!pcbs->set) return;
  myCh *pch = (myCh *) data;
  if (w == pch->_formatFixedButton) {
     pch->notation('f');
  } else
  if (w == pch->_formatScientificButton) {
     pch->notation('e');
  } else
  if (w == pch->_formatHexadecimalButton) {
     pch->notation('x');
  } 
  pch->valueChangedCb();
  if (pch->_infoDlg) {
    pch->updateInfoLimitsPanel();
    pch->updateInfoHistoryPanel();
  }
  return;
}


void myCh::adjustButtonsCb(Widget w, XtPointer data, XtPointer cbs) {
  myCh *pch = (myCh *) data;
  XmToggleButtonCallbackStruct *pcbs = (XmToggleButtonCallbackStruct *) cbs;

  // make change only on key pressed
  if (!pcbs->set) return;

  // check for which button is pressed
  // send the value;
  int i;
  for (i=0; i < pch->pchData->_info.e.no_str; i++) {
    if (w == pch->_adjustButtons[i]) {
      pch->pchData->self->send(*pch, (long) i);
      return;
    }
  }
}

void myCh::createButtonsPanel() {
  int i;
  // create a buttons panel to hold sixteen buttons
  // create a frame widget to outline the panel
  _adjustButtonPanelFrame = new vFrame(_adjustControlPanel,"State");
  _adjustButtonPanel = XtVaCreateManagedWidget("buttonPanel",
	    xmRowColumnWidgetClass, _adjustButtonPanelFrame->frame(),
	    XmNpacking, XmPACK_COLUMN,
	    XmNradioBehavior, TRUE,
	    XmNisHomogeneous, TRUE,
	    XmNindicatorType, XmONE_OF_MANY,
	    XmNvisibleWhenOff, TRUE,
	    XmNentryClass, xmToggleButtonWidgetClass,
	    XmNradioAlwaysOne, True,
	    XmNnumColumns, 4,
	    NULL);
  for (i=0; i<pchData->_info.e.no_str; i++) {
    if (strlen(pchData->_info.e.strs[i]) > 0) {
      _adjustButtons[i] = XtVaCreateManagedWidget(pchData->_info.e.strs[i],
	    xmToggleButtonWidgetClass,_adjustButtonPanel,
	    NULL);
    } else {
      _adjustButtons[i] = XtVaCreateManagedWidget(buttonLabels[i],
	    xmToggleButtonWidgetClass,_adjustButtonPanel,
	    NULL);
    }
    if (i == pchData->_info.e.value) {
      XtVaSetValues(_adjustButtons[i],
	    XmNset, TRUE,
	    NULL);
    }
    XtAddCallback(_adjustButtons[i],XmNvalueChangedCallback,
		  myCh::adjustButtonsCb,this);
  }
  _adjustButtonPanelFrame->manage();
}

void myCh::updateAdjustParameters() {
  switch (ca_field_type(pchData->_chID)) {
  case DBF_STRING :
    return;
  case DBF_ENUM :
    return;
  case DBF_CHAR :
    _adjustOpHigh = (double) pchData->_info.c.upper_ctrl_limit;
    _adjustOpLow = (double) pchData->_info.c.lower_ctrl_limit;
    if (_adjustOpHigh < _adjustOpLow) {
      _adjustOpHigh = (double) pchData->_info.c.lower_ctrl_limit;
      _adjustOpLow = (double) pchData->_info.c.upper_ctrl_limit;
    }
    _adjustStepSize = 1.0;
    break;
  case DBF_INT :
    _adjustOpHigh = (double) pchData->_info.i.upper_ctrl_limit;
    _adjustOpLow = (double) pchData->_info.i.lower_ctrl_limit;
    if (_adjustOpHigh < _adjustOpLow) {
      _adjustOpHigh = (double) pchData->_info.i.lower_ctrl_limit;
      _adjustOpLow = (double) pchData->_info.i.upper_ctrl_limit;
    }
    _adjustStepSize = 1.0;
    break;
  case DBF_LONG :
    _adjustOpHigh = (double) pchData->_info.l.upper_ctrl_limit;
    _adjustOpLow = (double) pchData->_info.l.lower_ctrl_limit;
    if (_adjustOpHigh < _adjustOpLow) {
      _adjustOpHigh = (double) pchData->_info.l.lower_ctrl_limit;
      _adjustOpLow = (double) pchData->_info.l.upper_ctrl_limit;
    }
    _adjustStepSize = 1.0;
    break;
  case DBF_FLOAT :
    _adjustOpHigh = pchData->_info.f.upper_ctrl_limit;
    _adjustOpLow = pchData->_info.f.lower_ctrl_limit;
    if (_adjustOpHigh < _adjustOpLow) {
      _adjustOpHigh = (double) pchData->_info.f.lower_ctrl_limit;
      _adjustOpLow = (double) pchData->_info.f.upper_ctrl_limit;
    }
    _adjustStepSize = pow(10,(-1*_precision));
    break;
  case DBF_DOUBLE :
    _adjustOpHigh = pchData->_info.d.upper_ctrl_limit;
    _adjustOpLow = pchData->_info.d.lower_ctrl_limit;
    if (_adjustOpHigh < _adjustOpLow) {
      _adjustOpHigh = (double) pchData->_info.d.lower_ctrl_limit;
      _adjustOpLow = (double) pchData->_info.d.upper_ctrl_limit;
    }
    _adjustStepSize = pow(10,(-1*_precision));
    break;
  default :
    cerr << "myCh::updateAdjustParameters : unknown type!" << endl;
    return;
  }
  if ((_adjustOpHigh == 0.0) && (_adjustOpLow == 0.0)) {
    _adjustOpHigh = 1.0;
  }
}

void myCh::setScrollBarMaxMin() {
  // the limits used by the scroll bar
  const double maxRange = 4000000.0;

  double pvOpRange;
  double pvValue;
  double sbOpRange;

  switch(ca_field_type(pchData->_chID)) {
  case DBF_STRING :
    cerr << "myCh::calcScrollBarMaxMin : string type encounted" << endl;
    return;
  case DBF_ENUM :
    cerr << "myCh::calcScrollBarMaxMin : binary type encounted" << endl;
    return;
  case DBF_CHAR :
    pvValue = (double) pchData->_pdata->c.value;
    break;
  case DBF_INT :
    pvValue = (double) pchData->_pdata->i.value;
    break;
  case DBF_LONG :
    pvValue = (double) pchData->_pdata->l.value;
    break;
  case DBF_FLOAT :
    pvValue = (double) pchData->_pdata->f.value;
    break;
  case DBF_DOUBLE :
    pvValue = pchData->_pdata->d.value;
    break;
  default :
    cerr << "myCh::calcScrollBarMaxMin : unknown type!" << endl;
    return;
  }
  if (pvValue > _adjustOpHigh) {
    _adjustOpHigh = pvValue;
  }
  if (pvValue < _adjustOpLow) {
    _adjustOpLow = pvValue;
  }
  pvOpRange = _adjustOpHigh - _adjustOpLow;
  sbOpRange = maxRange * _adjustStepSize;

  // now compare with the prcess variable range
  if (sbOpRange >= pvOpRange) {
    // it fits, the easiest case
    // find the center of the range and map it
    _adjustScrollBarOpHigh = _adjustOpHigh;
    _adjustScrollBarOpLow  = _adjustOpLow;
    _adjustScrollBarMax = (int) (pvOpRange  / _adjustStepSize);
    _adjustScrollBarMin = 0;
  } else {
    _adjustScrollBarOpHigh = pvValue + sbOpRange * 0.5;
    _adjustScrollBarOpLow  = _adjustScrollBarOpHigh - sbOpRange;
    if (_adjustScrollBarOpHigh > _adjustOpHigh) {
      _adjustScrollBarOpHigh = _adjustOpHigh;
      _adjustScrollBarOpLow = _adjustScrollBarOpHigh - sbOpRange;
    } else
    if (_adjustScrollBarOpLow < _adjustOpLow) {
      _adjustScrollBarOpLow = _adjustOpLow;
      _adjustScrollBarOpHigh = _adjustScrollBarOpLow + sbOpRange;
    }
    _adjustScrollBarMax = (int) maxRange;
    _adjustScrollBarMin = 0;
  } 
    
  _adjustScrollBarSliderSize = (_adjustScrollBarMax - _adjustScrollBarMin) / 10;
  if (_adjustScrollBarSliderSize < 1) _adjustScrollBarSliderSize = 1;
  XtVaSetValues(_adjustScrollBar,
	   XmNminimum, _adjustScrollBarMin,
	   XmNmaximum, _adjustScrollBarMax + _adjustScrollBarSliderSize,
	   XmNsliderSize, _adjustScrollBarSliderSize,
	   NULL);
  updateScrollBarMaxMinLabels();
  return;
}

void myCh::updateScrollBarMaxMinLabels() {
  switch(ca_field_type(pchData->_chID)) {
  case DBF_STRING :
    cerr << "myCh::updateScrollBarMaxMinLabels() : string type encounted!" << endl;
    return;
  case DBF_ENUM :
    cerr << "myCh::updateScrollBarMaxMinLabels() : enum type encounted!" << endl;
    return;
  case DBF_CHAR :
  case DBF_INT :
  case DBF_LONG :
    writeLabel(_adjustScrollBarHighLimitLabel,format((long)_adjustScrollBarOpHigh));
    writeLabel(_adjustScrollBarLowLimitLabel,format((long)_adjustScrollBarOpLow));
    break;
  case DBF_FLOAT:
  case DBF_DOUBLE :
    writeLabel(_adjustScrollBarHighLimitLabel,format(_adjustScrollBarOpHigh));
    writeLabel(_adjustScrollBarLowLimitLabel,format(_adjustScrollBarOpLow));
    break;
  default :
    cerr << "myCh::updateScrollBarMaxMinLabels() : unknown data type!!" << endl;
    return;
  }

  return;
}

void myCh::updateScrollBar() {
  int tmp;
  switch(ca_field_type(pchData->_chID)) {
  case DBF_STRING :
    cerr << "myCh::updateScrollBar() : string type encounted!" << endl;
    return;
  case DBF_ENUM :
    cerr << "myCh::updateScrollBar() : enum type encounted!" << endl;
    return;
  case DBF_CHAR :
    if ((pchData->_pdata->c.value > (unsigned char) _adjustScrollBarOpHigh) ||
        (pchData->_pdata->c.value < (unsigned char) _adjustScrollBarOpLow)) {
      setScrollBarMaxMin();
    }
    writeLabel(_adjustScrollBarValueLabel,stringValue());
    tmp = (int) ((double)pchData->_pdata->c.value - _adjustScrollBarOpLow);
    XtVaSetValues(_adjustScrollBar,XmNvalue,tmp,NULL);
    break;
  case DBF_INT :
    if ((pchData->_pdata->i.value > (short) _adjustScrollBarOpHigh) ||
        (pchData->_pdata->i.value < (short) _adjustScrollBarOpLow)) {
      setScrollBarMaxMin();
    }
    writeLabel(_adjustScrollBarValueLabel,stringValue());
    tmp = (int) ((double)pchData->_pdata->i.value - _adjustScrollBarOpLow);
    XtVaSetValues(_adjustScrollBar,XmNvalue,tmp,NULL);
    break;
  case DBF_LONG :
    if ((pchData->_pdata->l.value > (long) _adjustScrollBarOpHigh) ||
        (pchData->_pdata->l.value < (long) _adjustScrollBarOpLow)) {
      setScrollBarMaxMin();
    }
    writeLabel(_adjustScrollBarValueLabel,stringValue());
    tmp = (int) ((double)pchData->_pdata->l.value - _adjustScrollBarOpLow);
    XtVaSetValues(_adjustScrollBar,XmNvalue,tmp,NULL);
    break;
  case DBF_FLOAT:
    if ((pchData->_pdata->f.value > (float) _adjustScrollBarOpHigh) ||
        (pchData->_pdata->f.value < (float) _adjustScrollBarOpLow)) {
      setScrollBarMaxMin();
    }
    writeLabel(_adjustScrollBarValueLabel,stringValue());
    tmp = (int) ((pchData->_pdata->f.value - _adjustScrollBarOpLow)/_adjustStepSize);
    XtVaSetValues(_adjustScrollBar,XmNvalue,tmp,NULL);
    break;
  case DBF_DOUBLE :
    if ((pchData->_pdata->d.value > _adjustScrollBarOpHigh) ||
        (pchData->_pdata->d.value < _adjustScrollBarOpLow)) {
      setScrollBarMaxMin();
    }
    writeLabel(_adjustScrollBarValueLabel,stringValue());
    tmp = (int) ((pchData->_pdata->d.value - _adjustScrollBarOpLow)/_adjustStepSize);
    XtVaSetValues(_adjustScrollBar,XmNvalue,tmp,NULL);
    break;
  default :
    cerr << "myCh::updateScrollBar() : unknown data type!!" << endl;
    return;
  }
  return;
}

void myCh::adjustScrollBarCb(Widget, XtPointer data, XtPointer pcbs) {
  myCh *pch = (myCh *) data;
  XmScrollBarCallbackStruct *pscbs = (XmScrollBarCallbackStruct *) pcbs;
  if (!pch->connected()) return;
  switch (ca_field_type(pch->pchData->_chID)) {
  case DBF_STRING :
    cerr << "myCH::adjustScrollBarCb() : string type ecounted" << endl;
    return;
  case DBF_ENUM :
    cerr << "myCH::adjustScrollBarCb() : string enum ecounted" << endl;
    return;
  case DBF_CHAR :
  case DBF_INT :
  case DBF_LONG :
    {
      long tmp;
      tmp = (long) pscbs->value * (long) pch->_adjustStepSize +
	  (long) pch->_adjustScrollBarOpLow;
      pch->pchData->self->send(*pch,tmp);
    }
    break;
  case DBF_FLOAT :
  case DBF_DOUBLE :
    {
      double tmp;
      tmp = (double) pscbs->value * pch->_adjustStepSize +
	  pch->_adjustScrollBarOpLow;
      pch->pchData->self->send(*pch,tmp);
    }
    break;
  default :
    break;
  }
  return;
}

void myCh::createHorizontalScrollBarPanel() {
  // create a scrollBar panel to hold the scroll bar with
  // upper and lower limit range
  _adjustScrollBarPanelFrame = new vFrame(_adjustControlPanel,
					  "Valuator");
  _adjustScrollBarPanel = XtVaCreateManagedWidget("scrollBarPanel",
	   xmRowColumnWidgetClass, _adjustScrollBarPanelFrame->frame(),
	   NULL);
  // create a form widget to hold the value, low and high limit labels
  _adjustScrollBarLabelPanel = XtVaCreateManagedWidget("Label",
	   xmFormWidgetClass, _adjustScrollBarPanel,
	   XmNfractionBase, 5,
	   NULL);

  // create the value, low and high limit labels
  _adjustScrollBarLowLimitLabel = XtVaCreateManagedWidget("lower",
           xmLabelWidgetClass, _adjustScrollBarLabelPanel,
           XmNalignment, XmALIGNMENT_BEGINNING,
           XmNtopAttachment,        XmATTACH_FORM,
           XmNbottomAttachment,     XmATTACH_FORM,
           XmNleftAttachment,       XmATTACH_POSITION,
           XmNleftPosition,         0,
           XmNrightAttachment,      XmATTACH_POSITION,
           XmNrightPosition,        1,
           NULL);
  _adjustScrollBarValueLabel = XtVaCreateManagedWidget("value",
           xmLabelWidgetClass, _adjustScrollBarLabelPanel,
           XmNalignment, XmALIGNMENT_CENTER,
           XmNtopAttachment,        XmATTACH_FORM,
           XmNbottomAttachment,     XmATTACH_FORM,
           XmNleftAttachment,       XmATTACH_POSITION,
           XmNleftPosition,         2,
           XmNrightAttachment,      XmATTACH_POSITION,
           XmNrightPosition,        3,
           NULL);
  _adjustScrollBarHighLimitLabel = XtVaCreateManagedWidget("upper",
           xmLabelWidgetClass, _adjustScrollBarLabelPanel,
           XmNalignment, XmALIGNMENT_END,
           XmNtopAttachment,        XmATTACH_FORM,
           XmNbottomAttachment,     XmATTACH_FORM,
           XmNleftAttachment,       XmATTACH_POSITION,
           XmNleftPosition,         4,
           XmNrightAttachment,      XmATTACH_POSITION,
           XmNrightPosition,        5,
           NULL);
  // create the slider
  _adjustScrollBar = XtVaCreateManagedWidget("scrollBar",
	   xmScrollBarWidgetClass, _adjustScrollBarPanel,
	   XmNorientation, XmHORIZONTAL,
	   NULL);
  _adjustScrollBarPanelFrame->manage();
  setScrollBarMaxMin();
  updateScrollBar();
  XtAddCallback(_adjustScrollBar,XmNvalueChangedCallback,adjustScrollBarCb,this);
  XtAddCallback(_adjustScrollBar,XmNdragCallback,adjustScrollBarCb,this);
  return;
}

void myCh::adjustTextEntryCb(Widget w, XtPointer data, XtPointer) {
  myCh *pch = (myCh *) data;
  pch->pchData->self->send(*pch, XmTextFieldGetString(w));
  return;
}

void myCh::adjustCloseButtonCb(Widget, XtPointer data, XtPointer) {
  myCh *pch = (myCh *) data;
  
  if (pch->_adjustDlg) {
    delete pch->_adjustDlg;
    pch->_adjustDlg = NULL;
  }
  return;
}

void myCh::adjustCb(Widget widget, XtPointer data, XtPointer) {
  myCh *pch = (myCh *) data;
  Dimension h, w;
  Position x, y;

  if (pch->pchData == NULL) return;
  if (pch->pchData->_chID == NULL) return;
  // if no write access, let user changed it
  if (!ca_write_access(pch->pchData->_chID)) {
    return;
  }

  // see whether the widget is created before
  if (pch->_adjustDlg) {
    pch->_adjustDlg->manage();
    return;
  }

  // Makesure the channel is connected
  if (pch->connected() == False) return;

  // get dimension and position of the parent

  XtVaGetValues(topShell(widget),
	  XmNwidth, &w,
	  XmNheight, &h,
	  XmNx, &x,
	  XmNy, &y,
	  NULL);

  // create the dialog shell
  pch->_adjustDlg = new vDialog(widget,"adjust");
  pch->_adjustDlg->moveTo(x - 10, y + h + 15);

  // create the rowColumn widget to hold control elements
  pch->_adjustControlPanel = XtVaCreateManagedWidget("controlPanel",
	    xmRowColumnWidgetClass, pch->_adjustDlg->panel(),
	    NULL);

  // create a text entry
  pch->_adjustTextEntry = XtVaCreateManagedWidget("textEntry",
	    xmTextFieldWidgetClass, pch->_adjustControlPanel,
	    XmNcolumns, 40,
	    XmNeditable, TRUE,
	    NULL);
  XmTextFieldSetString(pch->_adjustTextEntry,pch->stringValue());
  XtAddCallback(pch->_adjustTextEntry, XmNactivateCallback,
		  myCh::adjustTextEntryCb, pch);

  /* Depends on the channel type, for analog value
   * I create a up down button for that, for 
   * state value I create a buttons panel
   */

  switch(ca_field_type(pch->pchData->_chID)) {
  case DBF_STRING :
    // do nothing
    break;
  case DBF_ENUM :
    pch->createButtonsPanel();
    break;
  case DBF_CHAR :
  case DBF_INT  :
  case DBF_LONG :
  case DBF_FLOAT :
  case DBF_DOUBLE :
    pch->createHorizontalScrollBarPanel();
    break;
  default :
    break;
  }

  // create the close button and its panel
  pch->_adjustActionPanel = XtVaCreateWidget("CloseButtonPanel",
	    xmFormWidgetClass, pch->_adjustDlg->panel(),
	    XmNfractionBase, 3,
	    NULL);

  pch->_adjustCloseButton = XtVaCreateManagedWidget("Close",
	    xmPushButtonWidgetClass, pch->_adjustActionPanel,
	    XmNtopAttachment,    XmATTACH_FORM,
	    XmNbottomAttachment, XmATTACH_FORM,
	    XmNleftAttachment,   XmATTACH_POSITION,
	    XmNleftPosition,     1,
	    XmNrightAttachment,  XmATTACH_POSITION,
	    XmNrightPosition,    2,
	    NULL);

  XtAddCallback(pch->_adjustCloseButton, XmNactivateCallback, 
		  myCh::adjustCloseButtonCb, pch);
  // fix the size of the action area
  XtManageChild(pch->_adjustActionPanel);
  XtVaGetValues(pch->_adjustCloseButton, XmNheight, &h, NULL);
  XtVaSetValues(pch->_adjustActionPanel,
	    XmNpaneMaximum, h,
	    XmNpaneMinimum, h,
	    NULL);

  pch->_adjustDlg->manage();
  return;
}

void myCh::infoCb(Widget widget, XtPointer data, XtPointer) {
  myCh *pch = (myCh *) data;
  Dimension h, w;
  Position x, y;

  // see whether the widget is created before
  if (pch->_infoDlg) {
    pch->_infoDlg->manage();
    return;
  }

  // Makesure the channel is connected
  if (pch->connected() == False) return;

  // get dimension and position of the parent

  XtVaGetValues(topShell(widget),
	  XmNwidth, &w,
	  XmNheight, &h,
	  XmNx, &x,
	  XmNy, &y,
	  NULL);

  // create the dialog shell
  pch->_infoDlg = new vDialog(widget, "info");
  pch->_infoDlg->moveTo( x + 15, y - 25);

  // create a rowColumn widget to hold the information
  pch->_infoDataPanel = XtVaCreateManagedWidget("dataPanel",
	    xmRowColumnWidgetClass, pch->_infoDlg->panel(),
	    NULL);

  // create the frame widget to hold limits label
  pch->_infoLimitsPanelFrame = new vFrame(pch->_infoDataPanel,
					  "PV's Information");

  // create a label widget to show limits
  if (pch->_font3) {
    pch->_infoLimitsLabel = XtVaCreateManagedWidget("limitsLabel",
	    xmLabelWidgetClass, pch->_infoLimitsPanelFrame->frame(),
	    XmNchildType, XmFRAME_WORKAREA_CHILD,
	    XmNalignment, XmALIGNMENT_BEGINNING,
	    XmNfontList, pch->_fontList3,
	    NULL);
  } else {
    pch->_infoLimitsLabel = XtVaCreateManagedWidget("limitsLabel",
	    xmLabelWidgetClass, pch->_infoLimitsPanelFrame->frame(),
	    XmNchildType, XmFRAME_WORKAREA_CHILD,
	    XmNalignment, XmALIGNMENT_BEGINNING,
	    NULL);
  }
  pch->_infoLimitsPanelFrame->manage();
  pch->updateInfoLimitsPanel();

  // create the frame widget to hold history label
  pch->_infoHistoryPanelFrame = new vFrame(pch->_infoDataPanel,
					   "History");

  // create a label widget to show history
  if (pch->_font3) {
    pch->_infoHistoryLabel = XtVaCreateManagedWidget("HistoryLabel",
	    xmLabelWidgetClass, pch->_infoHistoryPanelFrame->frame(),
	    XmNchildType, XmFRAME_WORKAREA_CHILD,
	    XmNalignment, XmALIGNMENT_BEGINNING,
	    XmNfontList, pch->_fontList3,
	    NULL);
  } else {
    pch->_infoHistoryLabel = XtVaCreateManagedWidget("HistoryLabel",
	    xmLabelWidgetClass, pch->_infoHistoryPanelFrame->frame(),
	    XmNchildType, XmFRAME_WORKAREA_CHILD,
	    XmNalignment, XmALIGNMENT_BEGINNING,
	    NULL);
  }

  pch->_infoHistoryPanelFrame->manage();
  pch->updateInfoHistoryPanel();

  // create the close button and its panel
  pch->_infoActionPanel = XtVaCreateWidget("CloseButtonPanel",
	    xmFormWidgetClass, pch->_infoDlg->panel(),
	    XmNfractionBase, 3,
	    NULL);

  pch->_infoCloseButton = XtVaCreateManagedWidget("Close",
	    xmPushButtonWidgetClass, pch->_infoActionPanel,
	    XmNtopAttachment,    XmATTACH_FORM,
	    XmNbottomAttachment, XmATTACH_FORM,
	    XmNleftAttachment,   XmATTACH_POSITION,
	    XmNleftPosition,     1,
	    XmNrightAttachment,  XmATTACH_POSITION,
	    XmNrightPosition,    2,
	    NULL);

  XtAddCallback(pch->_infoCloseButton, XmNactivateCallback, 
		  myCh::infoCloseButtonCb, pch);
  // fix the size of the action area
  XtManageChild(pch->_infoActionPanel);
  XtVaGetValues(pch->_infoCloseButton, XmNheight, &h, NULL);
  XtVaSetValues(pch->_infoActionPanel,
	    XmNpaneMaximum, h,
	    XmNpaneMinimum, h,
	    NULL);

  // fix the size of the panel and the shell
  pch->_infoDlg->manage();
  return;
}

void myCh::updateInfoLimitsPanel() {
  static char msg[1024];
  char recordType[80];
  ostrstream ostr(msg,1024);
  int w = 8;
  char *rt;
  if (pchData == NULL) {
    ostr << "No Information !" << ends;
  } else
  if (pchData->_state == NeverConnected) {
    ostr << "No Information !" << ends;
  } else
  if (pchData->_state == LostConnection) {
    // do nothing
    return;
  } else {

    *(this->RvCh) >> recordType;
    rt = recordType;

    switch(ca_field_type(pchData->_chID)) {
    case DBF_STRING :
      ostr << "Data Type     : String" << '\n';
      ostr << "Record Type   : " << rt << '\n';
      ostr << "Element Count : " << ca_element_count(pchData->_chID) << '\n';
      ostr << "Host name     : " << ca_host_name(pchData->_chID) << '\n';
      ostr << "Read Access   : " << (ca_read_access(pchData->_chID) ? "yes" : "no") << '\n';
      ostr << "Write Access  : " << (ca_write_access(pchData->_chID) ? "yes" : "no") << ends;
      break;
    case DBF_ENUM :
      ostr << "Data Type     : Enum" << '\n';
      ostr << "Record Type   : " <<  rt << '\n';
      ostr << "Element Count : " << ca_element_count(pchData->_chID) << '\n';
      ostr << "Host name     : " << ca_host_name(pchData->_chID) << '\n';
      ostr << "Read Access   : " << (ca_read_access(pchData->_chID) ? "yes" : "no") << '\n';
      ostr << "Write Access  : " << (ca_write_access(pchData->_chID) ? "yes" : "no") << ends;
      break;
    case DBF_CHAR :
      ostr << "Data type     : Character\n";
      ostr << "Record Type   : " <<  rt << '\n';
      ostr << "Units         : " << pchData->_info.c.units << '\n';
      ostr << "Element Count : " << ca_element_count(pchData->_chID) << '\n';
      ostr << "Host name     : " << ca_host_name(pchData->_chID) << '\n';
      ostr << "Read Access   : " << (ca_read_access(pchData->_chID) ? "yes" : "no") << '\n';
      ostr << "Write Access  : " << (ca_write_access(pchData->_chID) ? "yes" : "no") << '\n';
      ostr.flags(vCh::_format);
      ostr.setf(ios::right,ios::adjustfield);
      ostr << "Upper display limit (HOPR) : " << setw(w) << (int) pchData->_info.c.upper_disp_limit << '\n';
      ostr << "Lower display limit (LOPR) : " << setw(w) << (int) pchData->_info.c.lower_disp_limit << '\n';
      ostr << "Upper control limit (DRVH) : " << setw(w) << (int) pchData->_info.c.upper_ctrl_limit << '\n';
      ostr << "Lower control limit (DRVL) : " << setw(w) << (int) pchData->_info.c.lower_ctrl_limit << '\n';
      ostr << "Upper alarm limit   (HIHI) : " << setw(w) << (int) pchData->_info.c.upper_alarm_limit << '\n';
      ostr << "Lower alarm limit   (LOLO) : " << setw(w) << (int) pchData->_info.c.lower_alarm_limit << '\n';
      ostr << "Upper warning limit (HIGH) : " << setw(w) << (int) pchData->_info.c.upper_warning_limit << '\n';
      ostr << "Lower warning limit (LOW)  : " << setw(w) << (int) pchData->_info.c.lower_warning_limit << ends;
      break;
    case DBF_INT :
      ostr << "Data Type     : integer\n";
      ostr << "Record Type   : " << rt << '\n';
      ostr << "Units         : " << pchData->_info.i.units << '\n';
      ostr << "Element Count : " << ca_element_count(pchData->_chID) << '\n';
      ostr << "Host name     : " << ca_host_name(pchData->_chID) << '\n';
      ostr << "Read Access   : " << (ca_read_access(pchData->_chID) ? "yes" : "no") << '\n';
      ostr << "Write Access  : " << (ca_write_access(pchData->_chID) ? "yes" : "no") << '\n';
      ostr.flags(vCh::_format);
      ostr.setf(ios::right,ios::adjustfield);
      ostr << "Upper display limit (HOPR) : " << setw(w) << pchData->_info.i.upper_disp_limit << '\n';
      ostr << "Lower display limit (LOPR) : " << setw(w) << pchData->_info.i.lower_disp_limit << '\n';
      ostr << "Upper control limit (DRVH) : " << setw(w) << pchData->_info.i.upper_ctrl_limit << '\n';
      ostr << "Lower control limit (DRVL) : " << setw(w) << pchData->_info.i.lower_ctrl_limit << '\n';
      ostr << "Upper alarm limit   (HIHI) : " << setw(w) << pchData->_info.i.upper_alarm_limit << '\n';
      ostr << "Lower alarm limit   (LOLO) : " << setw(w) << pchData->_info.i.lower_alarm_limit << '\n';
      ostr << "Upper warning limit (HIGH) : " << setw(w) << pchData->_info.i.upper_warning_limit << '\n';
      ostr << "Lower warning limit (LOW)  : " << setw(w) << pchData->_info.i.lower_warning_limit << ends;
      break;
    case DBF_LONG :
      ostr << "Data Type     : long integer\n";
      ostr << "Record Type   : " <<  rt << '\n';
      ostr << "Units         : " << pchData->_info.l.units << '\n';
      ostr << "Element Count : " << ca_element_count(pchData->_chID) << '\n';
      ostr << "Host name     : " << ca_host_name(pchData->_chID) << '\n';
      ostr << "Read Access   : " << (ca_read_access(pchData->_chID) ? "yes" : "no") << '\n';
      ostr << "Write Access  : " << (ca_write_access(pchData->_chID) ? "yes" : "no") << '\n';
      ostr.flags(vCh::_format);
      ostr.setf(ios::right,ios::adjustfield);
      ostr.width(w);
      ostr << "Upper display limit (HOPR) : " << setw(w) << pchData->_info.l.upper_disp_limit << '\n';
      ostr << "Lower display limit (LOPR) : " << setw(w) << pchData->_info.l.lower_disp_limit << '\n';
      ostr << "Upper control limit (DRVH) : " << setw(w) << pchData->_info.l.upper_ctrl_limit << '\n';
      ostr << "Lower control limit (DRVL) : " << setw(w) << pchData->_info.l.lower_ctrl_limit << '\n';
      ostr << "Upper alarm limit   (HIHI) : " << setw(w) << pchData->_info.l.upper_alarm_limit << '\n';
      ostr << "Lower alarm limit   (LOLO) : " << setw(w) << pchData->_info.l.lower_alarm_limit << '\n';
      ostr << "Upper warning limit (HIGH) : " << setw(w) << pchData->_info.l.upper_warning_limit << '\n';
      ostr << "Lower warning limit (LOW)  : " << setw(w) << pchData->_info.l.lower_warning_limit << ends;
      break;
    case DBF_FLOAT :
      ostr << "Data Type     : single precision floating point number\n";
      ostr << "Record Type   : " <<  rt << '\n';
      ostr << "Units         : " << pchData->_info.f.units << '\n';
      ostr << "Precision     : " << pchData->_info.f.precision << '\n';
      ostr << "Element Count : " << ca_element_count(pchData->_chID) << '\n';
      ostr << "Host name     : " << ca_host_name(pchData->_chID) << '\n';
      ostr << "Read Access   : " << (ca_read_access(pchData->_chID) ? "yes" : "no") << '\n';
      ostr << "Write Access  : " << (ca_write_access(pchData->_chID) ? "yes" : "no") << '\n';
      ostr.precision(_precision);
      if (_notation == 'f') 
        w = w + _precision;
      ostr.flags(vCh::_format);
      ostr.setf(ios::right,ios::adjustfield);
      ostr << "Upper display limit (HOPR) : " << setw(w) << pchData->_info.f.upper_disp_limit << '\n';
      ostr << "Lower display limit (LOPR) : " << setw(w) << pchData->_info.f.lower_disp_limit << '\n';
      ostr << "Upper control limit (DRVH) : " << setw(w) << pchData->_info.f.upper_ctrl_limit << '\n';
      ostr << "Lower control limit (DRVL) : " << setw(w) << pchData->_info.f.lower_ctrl_limit << '\n';
      ostr << "Upper alarm limit   (HIHI) : " << setw(w) << pchData->_info.f.upper_alarm_limit << '\n';
      ostr << "Lower alarm limit   (LOLO) : " << setw(w) << pchData->_info.f.lower_alarm_limit << '\n';
      ostr << "Upper warning limit (HIGH) : " << setw(w) << pchData->_info.f.upper_warning_limit << '\n';
      ostr << "Lower warning limit (LOW)  : " << setw(w) << pchData->_info.f.lower_warning_limit << ends;
      break;
    case DBF_DOUBLE :
      ostr << "Data Type     : double precision floating point number\n";
      ostr << "Record Type   : " <<  rt << '\n';
      ostr << "Units         : " << pchData->_info.d.units << '\n';
      ostr << "Precision     : " << pchData->_info.d.precision << '\n';
      ostr << "Element Count : " << ca_element_count(pchData->_chID) << '\n';
      ostr << "Host name     : " << ca_host_name(pchData->_chID) << '\n';
      ostr << "Read Access   : " << (ca_read_access(pchData->_chID) ? "yes" : "no") << '\n';
      ostr << "Write Access  : " << (ca_write_access(pchData->_chID) ? "yes" : "no") << '\n';
      ostr.flags(vCh::_format);
      ostr.precision(_precision);
      if (_notation == 'f') 
        w = w + _precision;
      ostr.setf(ios::right,ios::adjustfield);
      ostr << "Upper display limit (HOPR) : " << setw(w) << pchData->_info.d.upper_disp_limit << '\n';
      ostr << "Lower display limit (LOPR) : " << setw(w) << pchData->_info.d.lower_disp_limit << '\n';
      ostr << "Upper control limit (DRVH) : " << setw(w) << pchData->_info.d.upper_ctrl_limit << '\n';
      ostr << "Lower control limit (DRVL) : " << setw(w) << pchData->_info.d.lower_ctrl_limit << '\n';
      ostr << "Upper alarm limit   (HIHI) : " << setw(w) << pchData->_info.d.upper_alarm_limit << '\n';
      ostr << "Lower alarm limit   (LOLO) : " << setw(w) << pchData->_info.d.lower_alarm_limit << '\n';
      ostr << "Upper warning limit (HIGH) : " << setw(w) << pchData->_info.d.upper_warning_limit << '\n';
      ostr << "Lower warning limit (LOW)  : " << setw(w) << pchData->_info.d.lower_warning_limit << ends;
      break;
    default :
      ostr << "Data Type : unknown" << ends;
      break;
    }
  }
  writeLabel(_infoLimitsLabel,msg);
  return;
}

void myCh::updateInfoHistoryPanel() {
  static char msg[512];
  ostrstream ostr(msg,512);
  char *tmp;
  int w = 8;
  if (pchData == NULL) {
    ostr << "No History !" << ends;
  } else
  if (pchData->_state == NeverConnected) {
    ostr << "No History !" << ends;
  } else
  if (pchData->_state == LostConnection) {
    // do nothing
    return;
  } else {
     switch(ca_field_type(pchData->_chID)) {
     case DBF_STRING : 
       {
         char timeStampStr[40];
	 char startTimeStr[40];
         struct tm *tblock = localtime(&_lastMinTime);
         strftime(timeStampStr,12,"%I:%M:%S %p",tblock);
         tblock = localtime(&_startTime);
         strftime(startTimeStr,12,"%I:%M:%S %p",tblock);
	 w = strlen(_minValue.s.value) + 3 + strlen(startTimeStr);
	 ostr << "Start Time  : " << setw(w) << startTimeStr << '\n';
         ostr << "Last String : " << _minValue.s.value;
         ostr << "   " << timeStampStr << ends;
       }
       break;
     case DBF_ENUM :
       {
         char timeStampStr[40];
	 char startTimeStr[40];
         struct tm *tblock = localtime(&_lastMinTime);
         strftime(timeStampStr,12,"%I:%M:%S %p",tblock);
         tblock = localtime(&_startTime);
         strftime(startTimeStr,12,"%I:%M:%S %p",tblock);
	 tmp = format(_minValue.e.value);
	 w = strlen(tmp)+3+strlen(startTimeStr);
	 ostr << "Start Time : " << setw(w) << startTimeStr << '\n';
         ostr << "Last State : " << format(_minValue.e.value); 
         ostr << "   " << timeStampStr << ends;
         break;
       }
     case DBF_CHAR :
       {
         char minTimeStampStr[40];
         char maxTimeStampStr[40];
	 char startTimeStr[40];
         struct tm *tblock = localtime(&_lastMinTime);
         strftime(minTimeStampStr,12,"%I:%M:%S %p",tblock);
         tblock = localtime(&_lastMaxTime);
         strftime(maxTimeStampStr,12,"%I:%M:%S %p",tblock);
         tblock = localtime(&_startTime);
         strftime(startTimeStr,12,"%I:%M:%S %p",tblock);
	 int w = 8;
	 ostr << "Start Time    : " << setw(w+13) << startTimeStr << '\n';
         ostr << "Maximum value : " << setw(w) << (int) _maxValue.c.value << "  " << maxTimeStampStr <<'\n';
         ostr << "Minimum value : " << setw(w) << (int) _minValue.c.value << "  " << minTimeStampStr << ends;
       }
       break;
     case DBF_INT  :
       {
         char minTimeStampStr[40];
         char maxTimeStampStr[40];
	 char startTimeStr[40];
         struct tm *tblock = localtime(&_lastMinTime);
         strftime(minTimeStampStr,12,"%I:%M:%S %p",tblock);
         tblock = localtime(&_lastMaxTime);
         strftime(maxTimeStampStr,12,"%I:%M:%S %p",tblock);
         tblock = localtime(&_startTime);
         strftime(startTimeStr,12,"%I:%M:%S %p",tblock);
	 int w = 8;
	 ostr << "Start Time    : " << setw(w+13) << startTimeStr << '\n';
         ostr << "Maximum value : " << setw(w) << _maxValue.i.value << "  " << maxTimeStampStr <<'\n';
         ostr << "Minimum value : " << setw(w) << _minValue.i.value << "  " << minTimeStampStr << ends;
       }
       break;
     case DBF_LONG :
       {
         char minTimeStampStr[40];
         char maxTimeStampStr[40];
	 char startTimeStr[40];
         struct tm *tblock = localtime(&_lastMinTime);
         strftime(minTimeStampStr,12,"%I:%M:%S %p",tblock);
         tblock = localtime(&_lastMaxTime);
         strftime(maxTimeStampStr,12,"%I:%M:%S %p",tblock);
         tblock = localtime(&_startTime);
         strftime(startTimeStr,12,"%I:%M:%S %p",tblock);
	 int w = 8;
	 ostr << "Start Time    : " << setw(w+13) << startTimeStr << '\n';
         ostr << "Maximum value : " << setw(w) << _maxValue.l.value << "  " << maxTimeStampStr <<'\n';
         ostr << "Minimum value : " << setw(w) << _minValue.l.value << "  " << minTimeStampStr << ends;
       }
       break;
     case DBF_FLOAT :
       {
         char minTimeStampStr[40];
         char maxTimeStampStr[40];
	 char startTimeStr[40];
         struct tm *tblock = localtime(&_lastMinTime);
         strftime(minTimeStampStr,12,"%I:%M:%S %p",tblock);
         tblock = localtime(&_lastMaxTime);
         strftime(maxTimeStampStr,12,"%I:%M:%S %p",tblock);
         tblock = localtime(&_startTime);
         strftime(startTimeStr,12,"%I:%M:%S %p",tblock);
	 int w = 15;
	 ostr.precision(_precision);
	 ostr.flags(vCh::_format);
         ostr.setf(ios::right,ios::adjustfield);
	 ostr << "Start Time    : " << setw(w+13) << startTimeStr << '\n';
         ostr << "Maximum value : " << setw(w) << _maxValue.f.value << "  " << maxTimeStampStr <<'\n';
         ostr << "Minimum value : " << setw(w) << _minValue.f.value << "  " << minTimeStampStr << ends;
       }
       break;
     case DBF_DOUBLE  :
       {
         char minTimeStampStr[40];
         char maxTimeStampStr[40];
	 char startTimeStr[40];
         struct tm *tblock = localtime(&_lastMinTime);
         strftime(minTimeStampStr,12,"%I:%M:%S %p",tblock);
         tblock = localtime(&_lastMaxTime);
         strftime(maxTimeStampStr,12,"%I:%M:%S %p",tblock);
         tblock = localtime(&_startTime);
         strftime(startTimeStr,12,"%I:%M:%S %p",tblock);
	 int w = 15;
	 ostr.precision(_precision);
	 ostr.flags(vCh::_format);
         ostr.setf(ios::right,ios::adjustfield);
	 ostr << "Start Time    : " << setw(w+13) << startTimeStr << '\n';
         ostr << "Maximum value : " << setw(w) << _maxValue.d.value << "  " << maxTimeStampStr <<'\n';
         ostr << "Minimum value : " << setw(w) << _minValue.d.value << "  " << minTimeStampStr << ends;
       }
       break;
     default :
       ostr << "Unknown Datatype !" << ends;
       break;
     }
  }

  writeLabel(_infoHistoryLabel,msg);

  return;
}

void myCh::infoCloseButtonCb(Widget , XtPointer data, XtPointer) {
  myCh *pch = (myCh *) data;
  if (pch->_infoDlg) {
    delete pch->_infoDlg;
    pch->_infoDlg = NULL;
  }
}

void myCh::createUserAndHostDlg(Widget widget, XtPointer data, XtPointer) {
  myCh *pch = (myCh *) data;
  Position x, y;
  Dimension w, h;
  
  // see whether the widget is created before
  if (pch->_userDlg == NULL) {

    // get the dimension and position of the widget
    XtVaGetValues(topShell(widget),
            XmNwidth, &w,
            XmNheight, &h,
	    XmNx, &x,
	    XmNy, &y,
            NULL);
   

    // create the dialog shell
    pch->_userDlg = new vDialog(widget,"access");
    pch->_userDlg->moveTo(x + w + 15, y - 25);

    // create a rowColumn widget to hold both host & user name text entries
    pch->_userDataPanel = 
	      XtVaCreateManagedWidget("dataPanel",
		  xmRowColumnWidgetClass,pch->_userDlg->panel(),
		  NULL);

    // create a frame widget to hold the user name text entries
    pch->_userNameEntryFrame = new vFrame(pch->_userDataPanel,
					  "User Name");

    // create a text widget for user name entry
    pch->_userNameEntry = 
	      XtVaCreateManagedWidget("UserName",
		  xmTextFieldWidgetClass,pch->_userNameEntryFrame->frame(),
		  XmNchildType, XmFRAME_WORKAREA_CHILD,
		  NULL);
    XtAddCallback(pch->_userNameEntry, XmNmodifyVerifyCallback,
		  myCh::userNameEntryCb, pch);

    pch->_userNameEntryFrame->manage();

    // create a frame widget to hold the host name text entries
    pch->_userHostEntryFrame = new vFrame(pch->_userDataPanel,
					  "Client Machine Name");

    // create a text widget for Host name entry
    pch->_userHostEntry = 
	      XtVaCreateManagedWidget("User Login Name",
		  xmTextFieldWidgetClass,pch->_userHostEntryFrame->frame(),
		  XmNchildType, XmFRAME_WORKAREA_CHILD,
		  NULL);
    XtAddCallback(pch->_userHostEntry, XmNmodifyVerifyCallback,
		  myCh::userHostEntryCb, pch);

    pch->_userHostEntryFrame->manage();

    // create the close button and its panel
    pch->_userActionPanel = XtVaCreateWidget("CloseButtonPanel",
	    xmFormWidgetClass, pch->_userDlg->panel(),
	    XmNfractionBase, 5,
	    NULL);

    pch->_userOkButton = XtVaCreateManagedWidget("Accept",
	    xmPushButtonWidgetClass, pch->_userActionPanel,
	    XmNtopAttachment,    XmATTACH_FORM,
	    XmNbottomAttachment, XmATTACH_FORM,
	    XmNleftAttachment,   XmATTACH_POSITION,
	    XmNleftPosition,     1,
	    XmNrightAttachment,  XmATTACH_POSITION,
	    XmNrightPosition,    2,
	    NULL);
    XtAddCallback(pch->_userOkButton, XmNactivateCallback,
		  myCh::userOkButtonCb, pch); 

    pch->_userCloseButton = XtVaCreateManagedWidget("Close",
	    xmPushButtonWidgetClass, pch->_userActionPanel,
	    XmNtopAttachment,    XmATTACH_FORM,
	    XmNbottomAttachment, XmATTACH_FORM,
	    XmNleftAttachment,   XmATTACH_POSITION,
	    XmNleftPosition,     3,
	    XmNrightAttachment,  XmATTACH_POSITION,
	    XmNrightPosition,    4,
	    NULL);
    XtAddCallback(pch->_userCloseButton, XmNactivateCallback,
		  myCh::userCloseButtonCb, pch); 

    // fix the size of the action area
    XtManageChild(pch->_userActionPanel);
    XtVaGetValues(pch->_userCloseButton, XmNheight, &h, NULL);
    XtVaSetValues(pch->_userActionPanel,
	    XmNpaneMaximum, h,
	    XmNpaneMinimum, h,
	    NULL);
  }

  // fix the size of the panel and the shell
  pch->_userDlg->manage();
}

void myCh::userCloseButtonCb(Widget, XtPointer data, XtPointer) {
  myCh *pch = (myCh *) data;
  
  if (pch->_userDlg) {
    delete pch->_userDlg;
    pch->_userDlg = NULL;
  }
  return;
}

void myCh::userOkButtonCb(Widget, XtPointer data, XtPointer) {
  myCh *pch = (myCh *) data;
  vCA  *pca = (vCA *) pch->pchData->self;
  if (pch->_userName) {
    if (strlen(pch->_userName)) {
      pca->userName(pch->_userName);
      // cerr << "user name = " << pch->_userName;
    }
  }
  if (pch->_hostName) {
    if (strlen(pch->_hostName)) {
      pca->hostName(pch->_hostName);
      // cerr << "host name = " << pch->_hostName;
    }
  }
  return;
}

void myCh::userNameEntryCb(Widget, XtPointer data, XtPointer cbs) {
  myCh *pch = (myCh *) data;
  XmTextVerifyCallbackStruct *pcbs = (XmTextVerifyCallbackStruct *) cbs;

  if ((pcbs->startPos == 0) &&
      (pcbs->startPos < pcbs->currInsert)) { // backspace 
    pcbs->endPos = strlen(pch->_userName);  // delete from here to end 
    pch->_userName[pcbs->startPos] = 0;    // backspace--terminated 
    return;
  }

  if (pcbs->text->length > 1) {
    pcbs->doit = False;            // don't allow "paste" operations 
    return;                       // make the user *type* password
  }

  char *tmp = new char[pcbs->endPos+2];  // new char + NULL terminator
  if (pch->_userName) {
    strcpy(tmp, pch->_userName);
    delete pch->_userName;
  } else 
    tmp[0] = '\0';

  pch->_userName = tmp;
  strncat(pch->_userName, pcbs->text->ptr, pcbs->text->length);
  pch->_userName[pcbs->endPos + pcbs->text->length] = '\0';

  int i;
  for (i=0; i < pcbs->text->length; i++)
    pcbs->text->ptr[i] = '*';
}

void myCh::userHostEntryCb(Widget, XtPointer data, XtPointer cbs) {
  myCh *pch = (myCh *) data;
  XmTextVerifyCallbackStruct *pcbs = (XmTextVerifyCallbackStruct *) cbs;

  if ((pcbs->startPos == 0) &&
      (pcbs->startPos < pcbs->currInsert)) { // backspace 
    pcbs->endPos = strlen(pch->_hostName);  // delete from here to end 
    pch->_hostName[pcbs->startPos] = 0;    // backspace--terminated 
    return;
  }

  if (pcbs->text->length > 1) {
    pcbs->doit = False;            // don't allow "paste" operations 
    return;                       // make the user *type* password
  }

  char *tmp = new char[pcbs->endPos+2];  // new char + NULL terminator
  if (pch->_hostName) {
    strcpy(tmp, pch->_hostName);
    delete pch->_hostName;
  } else 
    tmp[0] = '\0';

  pch->_hostName = tmp;
  strncat(pch->_hostName, pcbs->text->ptr, pcbs->text->length);
  pch->_hostName[pcbs->endPos + pcbs->text->length] = '\0';

  int i;
  for (i=0; i < pcbs->text->length; i++)
    pcbs->text->ptr[i] = '*';
}

extern "C" {
Widget createAndPopupProductDescriptionShell(
  XtAppContext appContext,       /* application context          */
  Widget topLevelShell,          /* application's topLevel shell */
  char *name,                    /* product/program name         */
  XmFontList nameFontList,       /*   and font list (or NULL)    */
  Pixmap namePixmap,             /* name Pixmap (or NULL)        */
  char *description,             /* product description          */
  XmFontList descriptionFontList,/*   and font list (or NULL)   */
  char *versionInfo,             /* product version number       */
  char *developedAt,             /* at and by...                 */
  XmFontList otherFontList,      /*   and font list (or NULL)    */
  int background,                /* background color (or -1)     */
  int foreground,                /* foreground color (or -1)     */
  int seconds);                   /* seconds to leave posted      */
}

void myCh::helpCb(Widget widget, XtPointer data, XtPointer) {
  myCh *pch = (myCh *) data;
  if (pch->_helpDlg) {
    XtPopup(pch->_helpDlg, XtGrabNone);
    return;
  }
  char versionString[50];
  ostrstream ostr(versionString,50);
  ostr << "PROBE VERSION SNS-1" << '(' << EPICS_VERSION_STRING << ')' << ends;
  // get to the top shell
  while (widget && !XtIsWMShell(widget))
    widget = XtParent(widget);

  if (widget == NULL) {
    cerr << "myCh::helpCb : parent is not a shell widget" << endl;
    return;
  }
  pch->_helpDlg = createAndPopupProductDescriptionShell(
    XtWidgetToApplicationContext(widget),widget,
    "Probe", pch->_fontList3, NULL,
    "Probe - a channel access utility", NULL,
    versionString,
    "developed at Argonne National Laboratory, by Frederick Vong", NULL,
    -1,-1,3);
  return;
}

