/*****************************************************************************
 *
 *     Author:	Frederick Vong
 *
 * Modification Log:
 * -----------------
 * .01	06-01-94	vong	initial version
 *
 ***************************************************************************/
/*
*****************************************************************
                          COPYRIGHT NOTIFICATION
*****************************************************************

THE FOLLOWING IS A NOTICE OF COPYRIGHT, AVAILABILITY OF THE CODE,
AND DISCLAIMER WHICH MUST BE INCLUDED IN THE PROLOGUE OF THE CODE
AND IN ALL SOURCE LISTINGS OF THE CODE.
 
(C)  COPYRIGHT 1993 UNIVERSITY OF CHICAGO
 
Argonne National Laboratory (ANL), with facilities in the States of 
Illinois and Idaho, is owned by the United States Government, and
operated by the University of Chicago under provision of a contract
with the Department of Energy.

Portions of this material resulted from work developed under a U.S.
Government contract and are subject to the following license:  For
a period of five years from March 30, 1993, the Government is
granted for itself and others acting on its behalf a paid-up,
nonexclusive, irrevocable worldwide license in this computer
software to reproduce, prepare derivative works, and perform
publicly and display publicly.  With the approval of DOE, this
period may be renewed for two additional five year periods. 
Following the expiration of this period or periods, the Government
is granted for itself and others acting on its behalf, a paid-up,
nonexclusive, irrevocable worldwide license in this computer
software to reproduce, prepare derivative works, distribute copies
to the public, perform publicly and display publicly, and to permit
others to do so.

*****************************************************************
                                DISCLAIMER
*****************************************************************

NEITHER THE UNITED STATES GOVERNMENT NOR ANY AGENCY THEREOF, NOR
THE UNIVERSITY OF CHICAGO, NOR ANY OF THEIR EMPLOYEES OR OFFICERS,
MAKES ANY WARRANTY, EXPRESS OR IMPLIED, OR ASSUMES ANY LEGAL
LIABILITY OR RESPONSIBILITY FOR THE ACCURACY, COMPLETENESS, OR
USEFULNESS OF ANY INFORMATION, APPARATUS, PRODUCT, OR PROCESS
DISCLOSED, OR REPRESENTS THAT ITS USE WOULD NOT INFRINGE PRIVATELY
OWNED RIGHTS.  

*****************************************************************
LICENSING INQUIRIES MAY BE DIRECTED TO THE INDUSTRIAL TECHNOLOGY
DEVELOPMENT CENTER AT ARGONNE NATIONAL LABORATORY (708-252-2000).
*/

#include "vCA.h"
#include <stdio.h>
extern int gethostname(char *, int);

vCA::vCA() {
  _count = 0;       // _count = 0, no channel is created
  _max = MAX_CONNECTION;

  // initialize the data structure
  chData *tmp = new chData[_max];
  int i;
  for (i=0;i<_max;i++) {
    tmp[i].self = this;
    tmp[i]._chID = NULL;
    tmp[i]._eventID = NULL;
    tmp[i]._state = NeverConnected;
    tmp[i]._pdata = NULL;
    tmp[i]._size = 0;
    pchList[i] = &(tmp[i]);
  }

  // start the channel access stuff
  _status = ca_task_initialize();
  SEVCHK(_status,"vCA::vCA() : ca_task_initialize failed!");
  if (_status != ECA_NORMAL) return;

  _status = ca_add_exception_event(exceptionHandlerCb, this);
  SEVCHK(_status,"vCA::vCA() : ca_add_exception_event failed!");
  if (_status != ECA_NORMAL) return;

  _status = ca_pend_io(0.1);

  return;
};

vCA::~vCA() {
  if (_count > 0) {
    // remove all the channel connection
    int i; 
    for (i=0;i<_count;i++) {
      chData *pdata = pchList[i];
      if (pdata->_eventID != NULL) {
        _status = ca_clear_event(pdata->_eventID);
        SEVCHK(_status,"vCA::~vCA() : ca_clear_event() failed!");
      }
      if (pdata->_chID != NULL) {
        _status = ca_clear_channel(pdata->_chID);
        SEVCHK(_status,"vCA::~vCA() : ca_clear_channel() failed!");
      }
    }
    _status = ca_pend_io(1.0);
    SEVCHK(_status,"vCA::~vCA() : ca_pend_io() failed!");
  }
  _status = ca_task_exit();
}

void vCA::connectionHandlerCb(struct connection_handler_args args) {
  chData *pdata = (chData *) ca_puser(args.chid);
  vCA    *pca = pdata->self;
  vCh    *pch = pdata->pch;

  switch (args.op) {
  case CA_OP_CONN_UP :

    /* If this is first time,
     * monitor the channel
     */

    if (pdata->_state == NeverConnected) {
      pdata->_state = ConnectionOK;
      pca->monitor(*pch);
      pca->_status = ca_replace_access_rights_event(pdata->_chID,
	  vCA::replaceAccessRightsEventCb);
    }

    /* I won't do the ch.connectionCb
     * until I get the range settings.
     */

    pdata->_state = ConnectionOK;
    if (ca_read_access(pdata->_chID)) {
      if (pca->getInfo(*pch) != ECA_NORMAL) {
        cerr << "vCA::connectionHandlerCb() : vCA::getInfo() failed!" << endl;
        cerr << "reason : " << ca_message(pca->_status) << endl;
      }
    } else {
      pch->connectCb();
    }
    break;
  case CA_OP_CONN_DOWN :
    pdata->_state = LostConnection;
    pch->disconnectCb();
    break;
  default :
    cerr << "vCA::connectionHandlerCb : unknown operation!" << endl;
    break;
  }
  return;
}

void vCA::exceptionHandlerCb(struct exception_handler_args args) {

  vCA *pca = (vCA *) args.usr;
  chid chID = args.chid;           // get channel ID
  pca->_status = (int) args.stat;  // get channel status

  // Find the channel which triggers the exception event
  int i;
  for (i=0;i<pca->_count;i++) {
    if (pca->pchList[i]->_chID == chID) break;
  }

  // Can't find the channel, bad happens
  if (i == pca->_count) {
    cerr << "vCA::exceptionHandlerCb : internal Error !!!" << endl;
    if (chID)
    	cerr << "Exceptional Callback channel : " << ca_name(chID) << endl;
    return;
  }

  // call the corresponding channel exception handler lists
  pca->pchList[i]->pch->exceptionEventCb();

  return;
}

void vCA::replaceAccessRightsEventCb(struct access_rights_handler_args args) {

  chData *pdata = (chData *) ca_puser(args.chid);
  vCA *pca = pdata->self;
  vCh *pch = pdata->pch;

  if (ca_state(pdata->_chID)!=cs_conn) return;
  if (ca_read_access(pdata->_chID)) {
    if (pca->getInfo(*pch) != ECA_NORMAL) {
      cerr << "vCA::connectionHandlerCb() : vCA::getInfo() failed!" << endl;
      cerr << "reason : " << ca_message(pca->_status) << endl;
    }
    pca->get(*pch);
  }
  pdata->pch->accessRightsChangedCb();
  return;
}

void vCA::updateChCb(struct event_handler_args args) {
  chData *pdata = (chData *) args.usr;
  pdata->self->_status = (int) args.status;

  // get the unix time
  time(&(pdata->_time));

  // copy new data
  int nBytes = dbr_size_n(args.type, args.count);

  if (pdata->_pdata == NULL) {
    pdata->_pdata = (dataBuf *) new char[nBytes];
  } else 
  if (pdata->_size < nBytes) {
    delete pdata->_pdata;
    pdata->_pdata = (dataBuf *) new char[nBytes];
    pdata->_size = nBytes;
  }

  while (nBytes-- > 0) 
    ((char *)pdata->_pdata)[nBytes] = ((char *)args.dbr)[nBytes];

  if ((pdata->_oldData.s.status != pdata->_pdata->s.status) ||
      (pdata->_oldData.s.severity != pdata->_pdata->s.severity)) {
    pdata->pch->alarmChangedCb();
  }
  pdata->pch->valueChangedCb();
  pdata->_oldData = *(pdata->_pdata);
  return;
}

void vCA::getInfoCb(struct event_handler_args args) {
  int nBytes;
  chData *pdata = (chData *) args.usr;
  char *tmp = (char *) &(pdata->_info);

  nBytes = dbr_size_n(args.type, args.count);
  while (nBytes-- > 0) 
    tmp[nBytes] = ((char *)args.dbr)[nBytes];
  vCh *pch = pdata->pch;
  if (pch->_formatOverride == False) {
    switch (ca_field_type(pdata->_chID)) {
    case DBF_STRING :
    case DBF_ENUM :
    case DBF_CHAR :
    case DBF_INT  :
    case DBF_LONG :
      pch->_precision = 0;
      break;
    case DBF_FLOAT :
    case DBF_DOUBLE :
      pch->_precision = pdata->_info.d.precision;
      break;
    default :
      break;
    }
  }
  pch->connectCb();
}

int vCA::poll(void) {
  _status = ca_pend_event(0.002);
  if ((_status != ECA_NORMAL) && (_status != ECA_TIMEOUT)) {
    return _status;
  }
  _status = ca_flush_io();
  return _status;
}

int vCA::add(vCh& ch) {
  _status = ECA_NORMAL;
  // see whether the ch is added before
  chData *pdata;
  if ((ch._caID < 0) || (ch._caID >= _max)) {
    // new ch
    if (_count == _max - 1) {
      cerr << "Maximum number of channel reached!" << endl;
      return -1;
    }
    pdata = pchList[_count];
    pdata->pch = &ch;
    ch._caID = _count;
    ch.pchData = pdata;
    _count++;
  } else {
    // disconnect the channel first
    pdata = ch.pchData;
    if (pdata->_eventID != NULL) {
      _status = ca_clear_event(pdata->_eventID);
      SEVCHK(_status,"vCA::add() : ca_clear_event() failed!");
    }
    if (pdata->_chID != NULL) {
      _status = ca_clear_channel(pdata->_chID);
      SEVCHK(_status,"vCA::add() : ca_clear_channel() failed!");
    }
    _status = ca_pend_io(0.001);
    SEVCHK(_status,"vCA::add() : ca_pend_io() failed!");
    pdata->_state = NeverConnected;
    pdata->_eventID = NULL;
    pdata->_chID = NULL;
  }
  if (iscntrl(ch._name[0])) return _status;
  _status = ca_build_and_connect(ch._name,TYPENOTCONN,0,
            &(pdata->_chID),NULL,&vCA::connectionHandlerCb,pdata);
  if (_status != ECA_NORMAL) return _status;
  _status = ca_pend_io(1.0);
  return _status;
}

int vCA::remove(vCh& ch) {

  _status = ECA_NORMAL;

  // invalid id, return
  if ((ch._caID < 0) || (ch._caID >= _count)) return _status;

  // delete the channel 
  chData *pdata = pchList[ch._caID];
  if (pdata->_eventID != NULL) {
    _status = ca_clear_event(pdata->_eventID);
    SEVCHK(_status,"vCA::remove() : ca_clear_event() failed!");
  }
  if (pdata->_chID != NULL) {
    _status = ca_clear_channel(pdata->_chID);
    SEVCHK(_status,"vCA::remove() : ca_clear_channel() failed!");
  }
  _status = ca_pend_io(0.001);
  SEVCHK(_status,"vCA::remove() : ca_pend_io() failed!");

  // clean up the node
  pdata->_eventID = NULL;
  pdata->_chID = NULL;
  pdata->_state = NeverConnected;
  pdata->pch = NULL;

  // take the last one the fill the gap
  // if it is not the last one
  _count--;
  if (ch._caID < _count) {
    pchList[ch._caID] = pchList[_count];
    pchList[_count]->pch->_caID = ch._caID;  // change to new ID
    pchList[_count] = pdata;
  }

  // clean up the rest
  ch._caID = -1;          // set an invalid ID
  ch.pchData = NULL;      // remove the hook

  return _status;
}

int vCA::send(vCh& ch, long num) {
  if (ch.pchData->_state != ConnectionOK) return -1;
  _status = ca_put(DBF_LONG, ch.pchData->_chID, (void *) &num);
  SEVCHK(_status,"vCA::send(vCh& ch) : ca_put for long value failed!");
  _status = ca_flush_io();
  SEVCHK(_status,"vCA::send(vCh& ch) : ca_flush_io() failed!");
  return _status;
}
  
int vCA::send(vCh& ch, double num) {
  if (ch.pchData->_state != ConnectionOK) return -1;
  _status = ca_put(DBF_DOUBLE, ch.pchData->_chID, (void *) &num);
  SEVCHK(_status,"vCA::send(vCh& ch) : ca_put for double value failed!");
  _status = ca_flush_io();
  SEVCHK(_status,"vCA::send(vCh& ch) : ca_flush_io() failed!");
  return _status;
}
  
int vCA::send(vCh& ch, char *str) {
  if (ch.pchData->_state != ConnectionOK) return -1;
  _status = ca_put(DBF_STRING, ch.pchData->_chID, (void *) str);
  SEVCHK(_status,"vCA::send(vCh& ch) : ca_put for string value failed!");
  _status = ca_flush_io();
  SEVCHK(_status,"vCA::send(vCh& ch) : ca_flush_io() failed!");
  return _status;
}
  
int vCA::send(vCh& ch) {
  if (ch.pchData->_state != ConnectionOK) return -1;
  switch(ca_field_type(ch.pchData->_chID)) {
  case DBF_STRING :
    _status = ca_put(DBF_STRING, ch.pchData->_chID, ch.pchData->_pdata->s.value);
    SEVCHK(_status,"vCA::send(vCh& ch) : ca_put for valueS failed!");
    break;
  case DBF_ENUM :
    _status = ca_put(DBF_ENUM, ch.pchData->_chID, &(ch.pchData->_pdata->e.value));
    SEVCHK(_status,"vCA::send(vCh& ch) : ca_put for valueE failed!");
    break;
  case DBF_CHAR :
    _status = ca_put(DBF_CHAR, ch.pchData->_chID, &(ch.pchData->_pdata->c.value));
    SEVCHK(_status,"vCA::send(vCh& ch) : ca_put for valueC failed!");
    break;
  case DBF_INT :
    _status = ca_put(DBF_INT, ch.pchData->_chID, &(ch.pchData->_pdata->i.value));
    SEVCHK(_status,"vCA::send(vCh& ch) : ca_put for valueI failed!");
    break;
  case DBF_LONG :
    _status = ca_put(DBF_LONG, ch.pchData->_chID, &(ch.pchData->_pdata->l.value));
    SEVCHK(_status,"vCA::send(vCh& ch) : ca_put for valueL failed!");
    break;
  case DBF_FLOAT :
    _status = ca_put(DBF_FLOAT, ch.pchData->_chID, &(ch.pchData->_pdata->f.value));
    SEVCHK(_status,"vCA::send(vCh& ch) : ca_put for valueF failed!");
    break;
  case DBF_DOUBLE :
    _status = ca_put(DBF_DOUBLE, ch.pchData->_chID, &(ch.pchData->_pdata->d.value));
    SEVCHK(_status,"vCA::send(vCh& ch) : ca_put for valueD failed!");
    break;
  default :
    cerr << "vCA::send(vCh& ch) :: Unknown data type!!!" << endl;
    break;
  }
  _status = ca_flush_io();
  SEVCHK(_status,"vCA::send(vCh& ch) : ca_flush_io() failed!");
  return _status;
}

int vCA::get(vCh& ch) {

  _status = -1;
  // if channel is not connected,
  // return.

  if (ch.connected() == False) return _status;
  _status = ca_array_get_callback(
	       dbf_type_to_DBR_TIME(ca_field_type(ch.pchData->_chID)),
	       ca_element_count(ch.pchData->_chID),
	       ch.pchData->_chID,
	       vCA::updateChCb,
	       (void *)ch.pchData);
  _status = ca_pend_io(1.0);
  if ((_status != ECA_NORMAL) && (_status != ECA_TIMEOUT)) return _status;
  ch.valueChangedCb();
  ch.alarmChangedCb();
  return _status;
}

int vCA::get() {
  int i;
  for (i=0;i<_count;i++) {
    vCh *ch = pchList[i]->pch;
    cerr << "vCA::get (i= " << i << "; name = " << ch->_name << ")" << endl;
    _status = get(*ch);
  }
  return _status;
}

int vCA::getInfo(vCh& ch) {
  chData *pdata = ch.pchData;
  _status = -1;

  // if channel is not connected,
  // return failed.

  if (ch.connected() == False) {
    return _status;
  }

  // get the closest native type.
  switch (ca_field_type(pdata->_chID)) {
  case DBF_STRING :
    // do a dummy call
    _status = ca_array_get_callback(
		 dbf_type_to_DBR_STS(ca_field_type(pdata->_chID)),
	         1,pdata->_chID,vCA::getInfoCb,pdata);
    break;
  case DBF_ENUM :
  case DBF_CHAR :
  case DBF_INT :
  case DBF_LONG :
  case DBF_FLOAT :
  case DBF_DOUBLE :
    _status = ca_array_get_callback(
		 dbf_type_to_DBR_CTRL(ca_field_type(pdata->_chID)),
	         1,pdata->_chID,vCA::getInfoCb,pdata);
    break;
  default :
    cerr << "vCA::get(vCh& ch) : unknown data type!" << endl;
    break;
  }
  if (_status != ECA_NORMAL) {
    cerr << "vCA::getInfo(vCh& ch) : ca_array_get_callback() failed!" << endl;
    cerr << "reason : " << ca_message(_status) << endl;
    return _status;
  }
  return _status;
}

int vCA::monitor(vCh &ch) {
  chData *pdata = ch.pchData;
  if (pdata->_eventID) return ECA_NORMAL;
  if (pdata->_state != ConnectionOK) {
    cerr << "vCA::monitor(vCh &ch) : channel not connected!" << endl;
    return -1;    // Don't know what to return 
  }
  _status = ca_add_array_event(
		dbf_type_to_DBR_TIME(ca_field_type(pdata->_chID)),
		ca_element_count(pdata->_chID), pdata->_chID,
		&vCA::updateChCb, pdata,
		0.0,0.0,0.0, &(pdata->_eventID));
  _status = ca_pend_io(1.0);
  return _status;
}

int  vCA::stopMonitor(vCh &ch) {
  chData *pdata = ch.pchData;
  if (pdata->_eventID == NULL) return ECA_NORMAL;
  _status = ca_clear_event(pdata->_eventID);
  if (_status != ECA_NORMAL) return _status;
  _status = ca_pend_io(1.0);
  if (_status != ECA_NORMAL) return _status;
  pdata->_eventID = NULL;
  return _status;
}

vCh::vCh(const char *name) {
  ostrstream ost(_name,PV_NAME_SIZE+1);
  ost << name << ends;
  _caID = -1;
  pchData = NULL;
  _access = False;
  _limitOverride = False;
  // init. formatStr
  _formatOverride = False;
  _precision = 0;
  _notation = 'f';
  _format = ios::left|ios::dec|ios::fixed;
  return;
}

int vCh::precision(int p) {
  _formatOverride = True;
  _precision = p;
  return _precision;
}


char *vCh::name(const char *newName) {
  ostrstream ost(_name,PV_NAME_SIZE+1);
  if (iscntrl(newName[0])) {
    ost << ends;
  } else {
    ost << newName << ends;
  }
  if (pchData != NULL)
    pchData->self->add(*this);
  return _name;
}

int vCh::notation(char n) {
  _formatOverride = True;
  _notation = n;
  switch(n) {
  case 'f' :
    _format = ios::left|ios::dec|ios::fixed;
    break;
  case 'e' :
    _format = ios::left|ios::dec|ios::scientific;
    break;
  case 'x' :
    _format = ios::left|ios::hex|ios::fixed|ios::showbase;
    break;
  default :
    break;
  }
  return (int) _notation;
}

ostream& operator << (ostream& os, const vCh& ch) {
  if (ch.pchData->_state != ConnectionOK) {
    os << "not connected!!!";
    return os;
  }
  if (!ca_read_access(ch.pchData->_chID)) {
    os << "";
    return os;
  }
  switch (ca_field_type(ch.pchData->_chID)) {
  case DBF_STRING :
    os << ch.pchData->_pdata->s.value;
    break;
  case DBF_ENUM :
    os.flags(ch._format);
    os << ch.pchData->_pdata->e.value;
    break;
  case DBF_CHAR :
    os << ch.pchData->_pdata->c.value;
    os.flags(ch._format);
    break;
  case DBF_INT :
    os << ch.pchData->_pdata->i.value;
    os.flags(ch._format);
    break;
  case DBF_LONG :
    os << ch.pchData->_pdata->l.value;
    os.flags(ch._format);
    break;
  case DBF_FLOAT :
    os.precision(ch._precision);
    os.flags(ch._format);
    os << ch.pchData->_pdata->f.value;
    break;
  case DBF_DOUBLE :
    os.precision(ch._precision);
    os.flags(ch._format);
    os << ch.pchData->_pdata->d.value;
    break;
  default :
    break;
  }
  return os;
}

istream& operator >> (istream& is, vCh& ch) {
  char buff[41];
  if (ch.pchData->_state != ConnectionOK) {
    return is;
  }
  switch (ca_field_type(ch.pchData->_chID)) {
  case DBF_STRING :
    is.get(ch.pchData->_pdata->s.value,41,'\n');
    break;
  case DBF_ENUM :
    is >> ch.pchData->_pdata->e.value;
    break;
  case DBF_CHAR :
    is >> ch.pchData->_pdata->c.value;
    break;
  case DBF_INT :
    is >> ch.pchData->_pdata->i.value;
    break;
  case DBF_LONG :
    is >> ch.pchData->_pdata->l.value;
    break;
  case DBF_FLOAT :
    is >> ch.pchData->_pdata->f.value;
    break;
  case DBF_DOUBLE :
    is >> ch.pchData->_pdata->d.value;
    break;
  default :
    break;
  }
  return is;
}

long vCh::longValue() {
  if (pchData->_state != ConnectionOK) {
    return 0;
  }
  switch (ca_field_type(pchData->_chID)) {
  case DBF_STRING :
    return (long) strlen(pchData->_pdata->s.value);
  case DBF_ENUM :
    return (long) pchData->_pdata->e.value;
  case DBF_CHAR :
    return (long) pchData->_pdata->c.value;
  case DBF_INT :
    return (long) pchData->_pdata->i.value;
  case DBF_LONG :
    return pchData->_pdata->l.value;
  case DBF_FLOAT :
    return (long) pchData->_pdata->f.value;
  case DBF_DOUBLE :
    return (long) pchData->_pdata->d.value;
  default :
    break;
  }
  return 0;
}

vCh& operator >> (vCh& ch, long& num) {
  num = ch.longValue();
  return ch;
}

double vCh::doubleValue() {
  if (pchData->_state != ConnectionOK) {
    return 0.0;
  }
  switch (ca_field_type(pchData->_chID)) {
  case DBF_STRING :
    return (double) strlen(pchData->_pdata->s.value);
  case DBF_ENUM :
    return (double) pchData->_pdata->e.value;
  case DBF_CHAR :
    return (double) pchData->_pdata->c.value;
  case DBF_INT :
    return (double) pchData->_pdata->i.value;
  case DBF_LONG :
    return (double) pchData->_pdata->l.value;
  case DBF_FLOAT :
    return pchData->_pdata->f.value;
  case DBF_DOUBLE :
    return pchData->_pdata->d.value;
  default :
    break;
  }
  return 0.0;
}

char *vCh::format(double num) {
  ostrstream ost(_tmpStr,MAX_STRING_SIZE);
  ost.flags(_format);
  ost.precision(_precision);
  ost << num << ends;
  return _tmpStr;
}

char *vCh::format(long num) {
  ostrstream ost(_tmpStr,MAX_STRING_SIZE);
  ost.flags(_format);
  ost << num << ends;
  return _tmpStr;
}

char *vCh::format(unsigned short num) {
  ostrstream ost(_tmpStr,MAX_STRING_SIZE);
  if (ca_field_type(pchData->_chID) == DBF_ENUM) {
    unsigned short len = strlen(pchData->_info.e.strs[pchData->_pdata->e.value]);
    if ((len > 0) && ((num >= 0) && (num < len))) {
      ost << pchData->_info.e.strs[num] << ends;
    } else {
      ost.flags(_format);
      ost << num << ends;
    }
  } else {
    ost.flags(_format);
    ost << num << ends;
  }
  return _tmpStr;
}


vCh& operator >> (vCh& ch, double& num) {
  num = ch.doubleValue();
  return ch;
}

char *vCh::stringValue() {
  ostrstream ost(_formatStr,MAX_STRING_SIZE);
  chData *pdata = pchData;

  if (pdata->_state == NeverConnected) {
    ost << "No Connection!" << ends;
    return _formatStr;
  }
  if (pdata->_state == LostConnection) {
    ost <<"Lost Connection!" << ends;
    return _formatStr;
  }
  if (!ca_read_access(pdata->_chID)) {
    ost << "" << ends;
    return _formatStr;
  }
  switch (ca_field_type(pdata->_chID)) {
  case DBF_STRING :
    ost << pdata->_pdata->s.value << ends;
    break;
  case DBF_ENUM :
    ost.flags(_format);
    if (strlen(pdata->_info.e.strs[pdata->_pdata->e.value]) > 0) {
      ost << pdata->_info.e.strs[pdata->_pdata->e.value] << ends;
    } else {
      ost << pdata->_pdata->e.value << ends;
    }
    break;
  case DBF_CHAR :
    ost.flags(_format);
    ost << (int) pdata->_pdata->c.value << ends;
    break;
  case DBF_INT :
    ost.flags(_format);
    ost << pdata->_pdata->i.value << ends;
    break;
  case DBF_LONG :
    ost.flags(_format);
    ost << pdata->_pdata->l.value << ends;
    break;
  case DBF_FLOAT :
    ost.precision(_precision);
    ost.flags(_format);
    ost << pdata->_pdata->f.value << ends;
    break;
  case DBF_DOUBLE :
    ost.precision(_precision);
    ost.flags(_format);
    ost << pdata->_pdata->d.value << ends;
    break;
  default :
    ost << ends;
    break;
  }
  return _formatStr;
}

vCh& operator >> (vCh& ch, char *text) {
  ostrstream ost(text,80);
  chData *pdata = ch.pchData;

  if (pdata->_state == NeverConnected) {
    ost << "No Connection!" << ends;
    return ch;
  }
  if (pdata->_state == LostConnection) {
    ost <<"Lost Connection!" << ends;
    return ch;
  }
  if (!ca_read_access(pdata->_chID)) {
    ost << "" << ends;
    return ch;
  }
  switch (ca_field_type(pdata->_chID)) {
  case DBF_STRING :
    ost << pdata->_pdata->s.value << ends;
    break;
  case DBF_ENUM :
    if (strlen(pdata->_info.e.strs[pdata->_pdata->e.value]) > 0) {
      ost << pdata->_info.e.strs[pdata->_pdata->e.value] << ends;
    } else {
      ost << pdata->_pdata->e.value << ends;
    }
    break;
  case DBF_CHAR :
    ost.flags(ch._format);
    ost << (int) pdata->_pdata->c.value  << ' ' << pdata->_info.c.units << ends;
    break;
  case DBF_INT :
    ost.flags(ch._format);
    ost << pdata->_pdata->i.value  << ' ' << pdata->_info.i.units << ends;
    break;
  case DBF_LONG :
    ost.flags(ch._format);
    ost << pdata->_pdata->l.value  << ' ' << pdata->_info.l.units << ends;
    break;
  case DBF_FLOAT :
    ost.precision(ch._precision);
    ost.flags(ch._format);
    ost << pdata->_pdata->f.value << ' ' <<  pdata->_info.f.units << ends;
    break;
  case DBF_DOUBLE :
    ost.precision(ch._precision);
    ost.flags(ch._format);
    ost << pdata->_pdata->d.value << ' ' <<  pdata->_info.d.units << ends;
    break;
  default :
    break;
  }
  return ch;
}

vCh& operator << (vCh& ch, const long& num) {
  if (ch.pchData->_state != ConnectionOK) {
    return ch;
  }
  switch (ca_field_type(ch.pchData->_chID)) {
  case DBF_STRING :
    sprintf(ch.pchData->_pdata->s.value,"%d",num);
    break;
  case DBF_ENUM :
    ch.pchData->_pdata->e.value = (short) num;
    break;
  case DBF_CHAR :
    ch.pchData->_pdata->c.value = (char) num;
    break;
  case DBF_INT :
    ch.pchData->_pdata->i.value = (short) num;
    break;
  case DBF_LONG :
    ch.pchData->_pdata->l.value = num;
    break;
  case DBF_FLOAT :
    ch.pchData->_pdata->f.value = (float) num;
    break;
  case DBF_DOUBLE :
    ch.pchData->_pdata->d.value = (double) num;
    break;
  default :
    break;
  }
  return ch;
}

vCh& operator << (vCh& ch, const double& num) {
  if (ch.pchData->_state != ConnectionOK) {
    return ch;
  }
  switch (ca_field_type(ch.pchData->_chID)) {
  case DBF_STRING :
    sprintf(ch.pchData->_pdata->s.value,"%f",num);
    break;
  case DBF_ENUM :
    ch.pchData->_pdata->e.value = (short) num;
    break;
  case DBF_CHAR :
    ch.pchData->_pdata->c.value = (char) num;
    break;
  case DBF_INT :
    ch.pchData->_pdata->i.value = (short) num;
    break;
  case DBF_LONG :
    ch.pchData->_pdata->l.value = (long) num;
    break;
  case DBF_FLOAT :
    ch.pchData->_pdata->f.value = (float) num;
    break;
  case DBF_DOUBLE :
    ch.pchData->_pdata->d.value = num;
    break;
  default :
    break;
  }
  return ch;
}

vCh& operator << (vCh& ch, char* text) {
  if (ch.pchData->_state != ConnectionOK) {
    return ch;
  }
  switch (ca_field_type(ch.pchData->_chID)) {
  case DBF_STRING :
    strcpy(ch.pchData->_pdata->s.value, text);
    break;
  case DBF_ENUM :
    sscanf(text,"%d",&(ch.pchData->_pdata->e.value));
    break;
  case DBF_CHAR :
    sscanf(text,"%d",&(ch.pchData->_pdata->c.value));
    break;
  case DBF_INT :
    sscanf(text,"%d",&(ch.pchData->_pdata->i.value));
    break;
  case DBF_LONG :
    sscanf(text,"%d",&(ch.pchData->_pdata->l.value));
    break;
  case DBF_FLOAT :
    sscanf(text,"%f",&(ch.pchData->_pdata->f.value));
    break;
  case DBF_DOUBLE :
    sscanf(text,"%f",&(ch.pchData->_pdata->d.value));
    break;
  default :
    break;
  }
  return ch;
}

void vCh::exceptionEventCb() {
  cerr << "CA exception handler entered for " << _name << endl;
  cerr << ca_message(pchData->self->status()) << endl;
  return;
}

int vCA::userName(char *name) {

  _status = ca_modify_user_name(name);
  SEVCHK(_status,"vCA::modifyUserName(char *) : ca_modifly_user_name() failed!");

  _status = ca_pend_io(0.001);

  return _status;
}

int vCA::hostName(char *name) {

  _status = ca_modify_host_name(name);
  SEVCHK(_status,"vCA::modifyHostName(char *) : ca_modifly_host_name() failed!");

  _status = ca_pend_io(0.001);

  return _status;
}
