/*****************************************************************************
 *
 *     Author:	Frederick Vong
 *
 * Modification Log:
 * -----------------
 * .01	06-01-94	vong	initial version
 *
 ***************************************************************************/
/*
*****************************************************************
                          COPYRIGHT NOTIFICATION
*****************************************************************

THE FOLLOWING IS A NOTICE OF COPYRIGHT, AVAILABILITY OF THE CODE,
AND DISCLAIMER WHICH MUST BE INCLUDED IN THE PROLOGUE OF THE CODE
AND IN ALL SOURCE LISTINGS OF THE CODE.
 
(C)  COPYRIGHT 1993 UNIVERSITY OF CHICAGO
 
Argonne National Laboratory (ANL), with facilities in the States of 
Illinois and Idaho, is owned by the United States Government, and
operated by the University of Chicago under provision of a contract
with the Department of Energy.

Portions of this material resulted from work developed under a U.S.
Government contract and are subject to the following license:  For
a period of five years from March 30, 1993, the Government is
granted for itself and others acting on its behalf a paid-up,
nonexclusive, irrevocable worldwide license in this computer
software to reproduce, prepare derivative works, and perform
publicly and display publicly.  With the approval of DOE, this
period may be renewed for two additional five year periods. 
Following the expiration of this period or periods, the Government
is granted for itself and others acting on its behalf, a paid-up,
nonexclusive, irrevocable worldwide license in this computer
software to reproduce, prepare derivative works, distribute copies
to the public, perform publicly and display publicly, and to permit
others to do so.

*****************************************************************
                                DISCLAIMER
*****************************************************************

NEITHER THE UNITED STATES GOVERNMENT NOR ANY AGENCY THEREOF, NOR
THE UNIVERSITY OF CHICAGO, NOR ANY OF THEIR EMPLOYEES OR OFFICERS,
MAKES ANY WARRANTY, EXPRESS OR IMPLIED, OR ASSUMES ANY LEGAL
LIABILITY OR RESPONSIBILITY FOR THE ACCURACY, COMPLETENESS, OR
USEFULNESS OF ANY INFORMATION, APPARATUS, PRODUCT, OR PROCESS
DISCLOSED, OR REPRESENTS THAT ITS USE WOULD NOT INFRINGE PRIVATELY
OWNED RIGHTS.  

*****************************************************************
LICENSING INQUIRIES MAY BE DIRECTED TO THE INDUSTRIAL TECHNOLOGY
DEVELOPMENT CENTER AT ARGONNE NATIONAL LABORATORY (708-252-2000).
*/

#ifndef VCA_H
#define VCA_H

#include <iostream.h>
#include <strstream.h>
#include <string.h>
#include <time.h>
#include <ctype.h>
extern "C" {
#include <cadef.h>
#include <db_access.h>
}

typedef union {
  struct dbr_time_string s;
  struct dbr_time_enum   e;
  struct dbr_time_char   c;
  struct dbr_time_short  i;
  struct dbr_time_long   l;
  struct dbr_time_float  f;
  struct dbr_time_double d;
} dataBuf;

typedef union {
  struct dbr_sts_string  s;
  struct dbr_ctrl_enum   e;
  struct dbr_ctrl_char   c;
  struct dbr_ctrl_short  i;
  struct dbr_ctrl_long   l;
  struct dbr_ctrl_float  f;
  struct dbr_ctrl_double d;
} infoBuf;

typedef enum {False, True} boolean;
typedef enum {NeverConnected, LostConnection, ConnectionOK} caState;

//#define MAX_STRING_SIZE 41
#define PV_NAME_SIZE 32

class vCh;
class vCA;

typedef struct {
  chid      _chID;
  evid      _eventID;
  dataBuf * _pdata;
  dataBuf   _oldData;
  long      _time;
  long      _size;
  infoBuf   _info;
  caState   _state;
  vCA       *self;  // point to channel access class itself
  vCh       *pch;   // point to the corresponding channel
} chData;

class vCh {
  friend class vCA;
protected:
  char     _name[PV_NAME_SIZE];

  // channel access info
  int      _caID;
  chData   *pchData;
  boolean  _access;

  //data display limit
  boolean  _limitOverride;
  double   _highLimit;
  double   _lowLimit;
  char     _units[MAX_STRING_SIZE];

  // data display format
  boolean  _formatOverride;
  char     _formatStr[MAX_STRING_SIZE];
  char     _tmpStr[MAX_STRING_SIZE];
  char     _notation;
  int      _precision;
  int      _format;

public:
  vCh(const char *name);
  virtual ~vCh() {};
  char *name(const char *newName);
  char *name()  { return _name; };
  int connected() { return (pchData->_state == ConnectionOK); };
  int precision(int p);
  int precision() { return _precision; };
  int notation(char n);
  int notation() { return (int) _notation; };
  int setDefault();
  friend ostream& operator << (ostream& os, const vCh& ch);
  friend istream& operator >> (istream& is, vCh& ch);
  friend vCh& operator >> (vCh& ch, long& num);
  friend vCh& operator >> (vCh& ch, double& num);
  friend vCh& operator >> (vCh& ch, char *text);
  friend vCh& operator << (vCh& ch, const long& num);
  friend vCh& operator << (vCh& ch, const double& num);
  friend vCh& operator << (vCh& ch, char* text);
  char * stringValue();
  long   longValue();
  double doubleValue();
  char * format(long);           // This handles the char, int, long stuff
  char * format(double);         // This handles the float, double
  char * format(unsigned short); // This handles the enum
  virtual void connectCb() {};
  virtual void disconnectCb() {};
  virtual void valueChangedCb() {};
  virtual void alarmChangedCb() {};
  virtual void exceptionEventCb();
  virtual void accessRightsChangedCb() {};
};

#define MAX_CONNECTION 256
class vCA {
protected:
  int _status;
  int _count;
  int _max;
  chData *pchList[MAX_CONNECTION];
  static void updateChCb(struct event_handler_args arg);
  static void connectionHandlerCb(struct connection_handler_args arg);
  static void exceptionHandlerCb(struct exception_handler_args arg);
  static void getInfoCb(struct event_handler_args arg);
  static void replaceAccessRightsEventCb(struct access_rights_handler_args);
  int monitor(vCh& ch);
  int stopMonitor(vCh& ch);
  int getInfo(vCh& ch);
public:
  vCA();
  ~vCA();
  int add(vCh& ch);
  int remove(vCh& ch);
  int status() { return _status; };
  int poll(void);
  int send(vCh&, long);
  int send(vCh&, double);
  int send(vCh&, char*);
  int send(vCh& ch);
  int send();
  int get(vCh& ch);
  int get();
  int userName(char *);
  int hostName(char *);
};

#endif
