/*****************************************************************************
 *
 *     Author:	Frederick Vong
 *
 * Modification Log:
 * -----------------
 * .01	06-01-94	vong	initial version
 *
 ***************************************************************************/
/*
*****************************************************************
                          COPYRIGHT NOTIFICATION
*****************************************************************

THE FOLLOWING IS A NOTICE OF COPYRIGHT, AVAILABILITY OF THE CODE,
AND DISCLAIMER WHICH MUST BE INCLUDED IN THE PROLOGUE OF THE CODE
AND IN ALL SOURCE LISTINGS OF THE CODE.
 
(C)  COPYRIGHT 1993 UNIVERSITY OF CHICAGO
 
Argonne National Laboratory (ANL), with facilities in the States of 
Illinois and Idaho, is owned by the United States Government, and
operated by the University of Chicago under provision of a contract
with the Department of Energy.

Portions of this material resulted from work developed under a U.S.
Government contract and are subject to the following license:  For
a period of five years from March 30, 1993, the Government is
granted for itself and others acting on its behalf a paid-up,
nonexclusive, irrevocable worldwide license in this computer
software to reproduce, prepare derivative works, and perform
publicly and display publicly.  With the approval of DOE, this
period may be renewed for two additional five year periods. 
Following the expiration of this period or periods, the Government
is granted for itself and others acting on its behalf, a paid-up,
nonexclusive, irrevocable worldwide license in this computer
software to reproduce, prepare derivative works, distribute copies
to the public, perform publicly and display publicly, and to permit
others to do so.

*****************************************************************
                                DISCLAIMER
*****************************************************************

NEITHER THE UNITED STATES GOVERNMENT NOR ANY AGENCY THEREOF, NOR
THE UNIVERSITY OF CHICAGO, NOR ANY OF THEIR EMPLOYEES OR OFFICERS,
MAKES ANY WARRANTY, EXPRESS OR IMPLIED, OR ASSUMES ANY LEGAL
LIABILITY OR RESPONSIBILITY FOR THE ACCURACY, COMPLETENESS, OR
USEFULNESS OF ANY INFORMATION, APPARATUS, PRODUCT, OR PROCESS
DISCLOSED, OR REPRESENTS THAT ITS USE WOULD NOT INFRINGE PRIVATELY
OWNED RIGHTS.  

*****************************************************************
LICENSING INQUIRIES MAY BE DIRECTED TO THE INDUSTRIAL TECHNOLOGY
DEVELOPMENT CENTER AT ARGONNE NATIONAL LABORATORY (708-252-2000).
*/

#include "xCA.h"

xCA::xCA(XtAppContext appContext):vCA() {
  _appContext = appContext;
  _inp = NULL;
  _maxInps = 0;
  _numInps = 0;
  _status = ca_add_fd_registration(xCA::registerCA,this);
  SEVCHK(_status, "xCA::xCA(XtAppContext appContext) : ca_add_fd_registration failed!");
  _status = ca_pend_io(0.1);
  xCA::caPeriodicPoll(this, NULL);
};

void xCA::registerCA(void *data, int fd, int condition)
{
  #define NUM_INITIAL_FDS 10

  int i, j;
  xCA *pca = (xCA *) data;

  if (pca->_inp == NULL && pca->_maxInps == 0) {
    /* first time through */
    pca->_inp = (InputIdAndFd *) new InputIdAndFd[NUM_INITIAL_FDS];
    pca->_maxInps = NUM_INITIAL_FDS;
    pca->_numInps = 0;
  }

  if (condition) {
    /* Add new fd */
    if (pca->_numInps < pca->_maxInps-1) {
      pca->_inp[pca->_numInps].fd = fd;
      pca->_inp[pca->_numInps].inputId = XtAppAddInput(pca->_appContext,fd,
                               (XtPointer) XtInputReadMask, xCA::caInputCb, pca);
      pca->_numInps++;
    } else {
      // allocate more memory 
      InputIdAndFd *tmp = (InputIdAndFd *) new InputIdAndFd[pca->_maxInps*2];
      int k;
      for (k=0;k<pca->_maxInps;k++) {
	tmp[k] = pca->_inp[k];
      }
      delete pca->_inp;
      pca->_inp = tmp;
      pca->_maxInps = 2*pca->_maxInps;
      pca->_inp[pca->_numInps].fd = fd;
      pca->_inp[pca->_numInps].inputId = XtAppAddInput(pca->_appContext,fd,
                             (XtPointer) XtInputReadMask, xCA::caInputCb , pca);
      pca->_numInps++;
     }
  } else {
    /* remove old fd/inputId */    
    for (i=0;i<pca->_numInps;i++) {
      if (pca->_inp[i].fd == fd) {
        XtRemoveInput(pca->_inp[i].inputId);
        pca->_inp[i].inputId = NULL;
        pca->_inp[i].fd = NULL;
      }
    }
    /* now remove holes in the array */
    i = 0;
    while (i < pca->_numInps) {
      if (pca->_inp[i].inputId == NULL) {
        j = i;
        pca->_numInps--;
        while (j < pca->_numInps) {
          pca->_inp[j] = pca->_inp[j+1];
          j++;
        }
      }
      i++;
    }
  }
}

void xCA::caInputCb(XtPointer data, int *, XtInputId *) {
  xCA *pca = (xCA *) data;
  pca->poll();
}

void xCA::caPeriodicPoll(XtPointer cd, XtIntervalId *) {
  xCA *pca = (xCA *) cd;
  pca->poll();
  XtAppAddTimeOut(pca->_appContext,100,xCA::caPeriodicPoll,pca);
}
